# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from awscli.autocomplete import filters, parser
from awscli.autocomplete.completer import CompletionResult
from awscli.autocomplete.local import basic, fetcher
from awscli.clidriver import CLIDriver, create_clidriver
from awscli.testutils import unittest
from tests.unit.autocomplete import InMemoryIndex


class TestShorthandCompleter(unittest.TestCase):
    def setUp(self):
        cli_driver = create_clidriver()
        self.cli_fetcher = fetcher.CliDriverFetcher(cli_driver)
        self.index = InMemoryIndex(
            {
                'command_names': {
                    '': [('aws', None)],
                    'aws': [
                        ('codebuild', None),
                        ('dynamodb', None),
                        ('s3', None),
                        ('ec2', None),
                        ('cloudformation', None),
                    ],
                    'aws.codebuild': [('create-project', None)],
                    'aws.dynamodb': [('put-item', None)],
                    'aws.ec2': [('bundle-instance', None)],
                    'aws.cloudformation': [('deploy', None)],
                },
                'arg_names': {
                    '': {
                        'aws': ['region', 'output', 'debug'],
                    },
                    'aws.codebuild': {
                        'create-project': ['source', 'secondary-sources'],
                    },
                    'aws.dynamodb': {
                        'put-item': ['item'],
                    },
                    'aws.ec2': {'bundle-instance': ['storage']},
                    'aws.cloudformation': {'deploy': ['capabilities']},
                },
                'arg_data': {
                    '': {
                        'aws': {
                            'region': (
                                'region',
                                'string',
                                'aws',
                                '',
                                None,
                                False,
                                False,
                            ),
                            'output': (
                                'output',
                                'string',
                                'aws',
                                '',
                                None,
                                False,
                                False,
                            ),
                            'debug': (
                                'debug',
                                'string',
                                'aws',
                                '',
                                None,
                                False,
                                False,
                            ),
                        }
                    },
                    'aws.codebuild': {
                        'create-project': {
                            'source': (
                                'source',
                                'structure',
                                'create-project',
                                'aws.codebuild.',
                                None,
                                False,
                                False,
                            ),
                            'secondary-sources': (
                                'secondary-sources',
                                'structure',
                                'create-project',
                                'aws.codebuild.',
                                None,
                                False,
                                False,
                            ),
                        }
                    },
                    'aws.dynamodb': {
                        'put-item': {
                            'item': (
                                'item',
                                'map',
                                'put-item',
                                'aws.dynamodb.',
                                None,
                                False,
                                False,
                            ),
                        }
                    },
                    'aws.ec2': {
                        'bundle-instance': {
                            'storage': (
                                'storage',
                                'structure',
                                'bundle-instance',
                                'aws.ec2.',
                                None,
                                False,
                                False,
                            ),
                        }
                    },
                    'aws.cloudformation': {
                        'deploy': {
                            'capabilities': (
                                'capabilities',
                                'list',
                                'deploy',
                                'aws.cloudformation.',
                                None,
                                False,
                                False,
                            ),
                        }
                    },
                },
            }
        )
        self.parser = parser.CLIParser(self.index)
        self.completer = basic.ShorthandCompleter(self.cli_fetcher)

    def assert_command_generates_suggestions(
        self, command_line, expected_suggestions
    ):
        parsed = self.parser.parse(command_line)
        suggestions = self.completer.complete(parsed)
        if expected_suggestions is None:
            self.assertIsNone(suggestions)
        else:
            displayed_suggestions = [x.display_text for x in suggestions]
            self.assertTrue(
                all(
                    [
                        suggest in displayed_suggestions
                        for suggest in expected_suggestions
                    ]
                ),
                command_line,
            )

    def test_return_none_if_it_does_not_have_shorthand_input(self):
        self.assert_command_generates_suggestions(
            'aws codebuild create-project --region', None
        )

    def test_return_suggestions_for_codebuild(self):
        self.assert_command_generates_suggestions(
            'aws codebuild create-project --source ',
            expected_suggestions=[
                'type=',
                'location=',
                'gitCloneDepth=',
                'gitSubmodulesConfig={',
                'buildspec=',
                'auth={',
                'reportBuildStatus=',
                'buildStatusConfig={',
                'insecureSsl=',
                'sourceIdentifier=',
            ],
        )

        self.assert_command_generates_suggestions(
            'aws codebuild create-project --source foo --secondary-sources ',
            expected_suggestions=[
                'type=',
                'location=',
                'gitCloneDepth=',
                'gitSubmodulesConfig={',
                'buildspec=',
                'auth={',
                'reportBuildStatus=',
                'buildStatusConfig={',
                'insecureSsl=',
                'sourceIdentifier=',
            ],
        )

        self.assert_command_generates_suggestions(
            'aws codebuild create-project --source bui',
            expected_suggestions=['buildspec=', 'buildStatusConfig={'],
        )

        self.assert_command_generates_suggestions(
            'aws codebuild create-project --source auth=',
            expected_suggestions=None,
        )

        self.assert_command_generates_suggestions(
            'aws codebuild create-project --source auth={',
            expected_suggestions=['type=', 'resource=', 'Autoclose brackets'],
        )

        self.assert_command_generates_suggestions(
            'aws codebuild create-project --source auth={type=',
            expected_suggestions=[
                'OAUTH',
                'CODECONNECTIONS',
                'SECRETS_MANAGER',
                'Autoclose brackets',
            ],
        )

        self.assert_command_generates_suggestions(
            'aws codebuild create-project '
            '--source auth={type=a},reportBuildStatus=',
            expected_suggestions=['true', 'false'],
        )

        self.assert_command_generates_suggestions(
            'aws codebuild create-project '
            '--source auth={type=a},reportBuildStatus=t',
            expected_suggestions=['true'],
        )

        self.assert_command_generates_suggestions(
            'aws codebuild create-project '
            '--source auth={type=a},reportBuildStatus=q',
            expected_suggestions=None,
        )

        self.assert_command_generates_suggestions(
            'aws codebuild create-project --source auth={type=a},location=',
            expected_suggestions=None,
        )

        self.assert_command_generates_suggestions(
            'aws codebuild create-project '
            '--source auth={type=a},location=foo,ty=',
            expected_suggestions=None,
        )

        self.assert_command_generates_suggestions(
            'aws codebuild create-project '
            '--source auth={type=a},location="foo,ty',
            expected_suggestions=None,
        )

        self.assert_command_generates_suggestions(
            'aws codebuild create-project '
            '--source auth={type=a},location=foo,type=',
            expected_suggestions=[
                'CODECOMMIT',
                'CODEPIPELINE',
                'GITHUB',
                'S3',
                'BITBUCKET',
                'GITHUB_ENTERPRISE',
                'NO_SOURCE',
                'GITLAB',
                'GITLAB_SELF_MANAGED',
            ],
        )

        self.assert_command_generates_suggestions(
            'aws codebuild create-project '
            '--source auth={type=a},location=foo,type=CO',
            expected_suggestions=['CODECOMMIT', 'CODEPIPELINE'],
        )

        self.assert_command_generates_suggestions(
            'aws codebuild create-project --source buildspec=foo,aut',
            expected_suggestions=['auth={'],
        )

    def test_return_suggestions_for_dynamodb(self):
        self.assert_command_generates_suggestions(
            'aws dynamodb put-item --item ', expected_suggestions=None
        )

        self.assert_command_generates_suggestions(
            'aws dynamodb put-item --item key={',
            expected_suggestions=[
                'S=',
                'N=',
                'B=',
                'SS=[',
                'NS=[',
                'BS=[',
                'M={',
                'L=[',
                'NULL=',
                'BOOL=',
                'Autoclose brackets',
            ],
        )

        self.assert_command_generates_suggestions(
            'aws dynamodb put-item --item key={SS',
            expected_suggestions=['SS=[', 'Autoclose brackets'],
        )

        self.assert_command_generates_suggestions(
            'aws dynamodb put-item --item key={M={',
            expected_suggestions=['Autoclose brackets'],
        )

        self.assert_command_generates_suggestions(
            'aws dynamodb put-item --item key={M={key1={',
            expected_suggestions=[
                'S=',
                'N=',
                'B=',
                'SS=[',
                'NS=[',
                'BS=[',
                'M={',
                'L=[',
                'NULL=',
                'BOOL=',
                'Autoclose brackets',
            ],
        )

        self.assert_command_generates_suggestions(
            'aws dynamodb put-item --item key={M={key1={M={key2={',
            expected_suggestions=[
                'S=',
                'N=',
                'B=',
                'SS=[',
                'NS=[',
                'BS=[',
                'M={',
                'L=[',
                'NULL=',
                'BOOL=',
                'Autoclose brackets',
            ],
        )

        self.assert_command_generates_suggestions(
            'aws dynamodb put-item --item key={M={key1={M={key2={B',
            expected_suggestions=['B=', 'BS=[', 'BOOL=', 'Autoclose brackets'],
        )

        self.assert_command_generates_suggestions(
            'aws dynamodb put-item --item key={BS=[1,2',
            expected_suggestions=['Autoclose brackets'],
        )

        self.assert_command_generates_suggestions(
            'aws dynamodb put-item --item key={BS=[1,2],',
            expected_suggestions=[
                'S=',
                'N=',
                'B=',
                'SS=[',
                'NS=[',
                'M={',
                'L=[',
                'NULL=',
                'BOOL=',
                'Autoclose brackets',
            ],
        )

        self.assert_command_generates_suggestions(
            'aws dynamodb put-item --item key={SS=[]}',
            expected_suggestions=None,
        )

    def test_autoclose_brackets(self):
        parsed = self.parser.parse(
            'aws dynamodb put-item --item ' 'key={M={key1={M={key2={BS=[1'
        )
        suggestions = self.completer.complete(parsed)
        self.assertEqual(
            suggestions[-1].name, 'key={M={key1={M={key2={BS=[1]}}}}}'
        )

        parsed = self.parser.parse(
            'aws dynamodb put-item --item ' 'key={M=[key1={M=[key2={BS=[1'
        )
        suggestions = self.completer.complete(parsed)
        self.assertEqual(
            suggestions[-1].name, 'key={M=[key1={M=[key2={BS=[1]}]}]}'
        )

    def test_not_start_autocompletion_wo_trailing_space(self):
        parsed = self.parser.parse('aws ec2 bundle-instance --storage')
        suggestions = self.completer.complete(parsed)
        self.assertIsNone(suggestions)

    def test_un_closable_brackets(self):
        parsed = self.parser.parse(
            'aws dynamodb put-item --item ' 'key={M={key1={M={key2=]{BS=[1'
        )
        suggestions = self.completer.complete(parsed)
        self.assertIsNone(suggestions)

    def test_fuzzy_search(self):
        self.completer = basic.ShorthandCompleter(
            self.cli_fetcher, response_filter=filters.fuzzy_filter
        )
        self.assert_command_generates_suggestions(
            'aws codebuild create-project --source auth={type=a},'
            'location=foo,type=OI',
            ['CODEPIPELINE', 'CODECOMMIT'],
        )

    def test_names_with_fuzzy_search(self):
        self.completer = basic.ShorthandCompleter(
            self.cli_fetcher, response_filter=filters.fuzzy_filter
        )
        parsed = self.parser.parse(
            'aws codebuild create-project --source buildspec=foo,at'
        )
        suggestions = self.completer.complete(parsed)
        names = [s.name for s in suggestions]
        display_text = [s.display_text for s in suggestions]
        self.assertEqual(
            names,
            [
                'buildspec=foo,location=',
                'buildspec=foo,buildStatusConfig={',
                'buildspec=foo,reportBuildStatus=',
                'buildspec=foo,auth={',
            ],
        )
        self.assertEqual(
            display_text,
            [
                'location=',
                'buildStatusConfig={',
                'reportBuildStatus=',
                'auth={',
            ],
        )

    def test_names_with_startswith_search(self):
        self.completer = basic.ShorthandCompleter(
            self.cli_fetcher, response_filter=filters.startswith_filter
        )
        parsed = self.parser.parse(
            'aws codebuild create-project --source auth={},bui'
        )
        suggestions = self.completer.complete(parsed)
        names = [s.name for s in suggestions]
        display_text = [s.display_text for s in suggestions]
        self.assertEqual(
            names, ['auth={},buildspec=', 'auth={},buildStatusConfig={']
        )
        self.assertEqual(display_text, ['buildspec=', 'buildStatusConfig={'])

    def test_return_suggestions_for_global_arg_with_choices(self):
        self.completer = basic.ShorthandCompleter(
            self.cli_fetcher, response_filter=filters.startswith_filter
        )
        parsed = self.parser.parse('aws --output ')
        suggestions = self.completer.complete(parsed)
        names = [s.name for s in suggestions]
        self.assertEqual(
            names, ['json', 'text', 'table', 'yaml', 'yaml-stream', 'off']
        )

    def test_not_return_suggestions_for_global_arg_wo_trailing_space(self):
        self.completer = basic.ShorthandCompleter(
            self.cli_fetcher, response_filter=filters.startswith_filter
        )
        parsed = self.parser.parse('aws --output')
        suggestions = self.completer.complete(parsed)
        self.assertIsNone(suggestions)

    def test_not_return_suggestions_for_global_arg_wo_choices(self):
        self.completer = basic.ShorthandCompleter(
            self.cli_fetcher, response_filter=filters.startswith_filter
        )
        parsed = self.parser.parse('aws --debug ')
        suggestions = self.completer.complete(parsed)
        self.assertIsNone(suggestions)

    def test_return_suggestions_for_list_of_enum(self):
        self.completer = basic.ShorthandCompleter(
            self.cli_fetcher, response_filter=filters.fuzzy_filter
        )
        parsed = self.parser.parse('aws cloudformation deploy --capabilities ')
        suggestions = self.completer.complete(parsed)
        self.assertIn(
            CompletionResult(
                'CAPABILITY_IAM', 0, False, None, None, 'CAPABILITY_IAM'
            ),
            suggestions,
        )
        self.assertIn(
            CompletionResult(
                'CAPABILITY_NAMED_IAM',
                0,
                False,
                None,
                None,
                'CAPABILITY_NAMED_IAM',
            ),
            suggestions,
        )

    def test_return_suggestions_for_list_of_enum_with_prefix(self):
        self.completer = basic.ShorthandCompleter(
            self.cli_fetcher, response_filter=filters.fuzzy_filter
        )
        parsed = self.parser.parse(
            'aws cloudformation deploy --capabilities ca'
        )
        suggestions = self.completer.complete(parsed)
        self.assertIn(
            CompletionResult(
                'CAPABILITY_IAM', 0, False, None, None, 'CAPABILITY_IAM'
            ),
            suggestions,
        )
        self.assertIn(
            CompletionResult(
                'CAPABILITY_NAMED_IAM',
                0,
                False,
                None,
                None,
                'CAPABILITY_NAMED_IAM',
            ),
            suggestions,
        )


class TestModelIndexCompleter(unittest.TestCase):
    def setUp(self):
        cli_driver = CLIDriver()
        self.cli_fetcher = fetcher.CliDriverFetcher(cli_driver)
        self.index = InMemoryIndex(
            {
                'command_names': {
                    '': [('aws', None)],
                },
                'arg_names': {
                    '': {
                        'aws': ['region', 'endpoint-url'],
                    },
                },
                'arg_data': {
                    '': {
                        'aws': {
                            'endpoint-url': (
                                'endpoint-url',
                                'string',
                                'aws',
                                '',
                                None,
                                False,
                                False,
                            ),
                            'region': (
                                'region',
                                'string',
                                'aws',
                                '',
                                None,
                                False,
                                False,
                            ),
                        }
                    }
                },
            }
        )
        self.parser = parser.CLIParser(self.index)
        self.completer = basic.ModelIndexCompleter(
            self.index, cli_driver_fetcher=self.cli_fetcher
        )

    def test_returns_help_text_for_params_in_global_scope(self):
        parsed = self.parser.parse('aws --re')
        suggestions = self.completer.complete(parsed)
        self.assertIn('The region', suggestions[0].help_text)


class TestQueryCompleter(unittest.TestCase):
    def setUp(self):
        cli_driver = CLIDriver()
        self.cli_fetcher = fetcher.CliDriverFetcher(cli_driver)
        index = InMemoryIndex(
            {
                'command_names': {
                    '': [('aws', None)],
                    'aws': [('ec2', None)],
                    'aws.ec2': [('describe-instances', None)],
                },
                'arg_names': {
                    '': {
                        'aws': ['query', 'output'],
                    },
                    'aws.ec2': {
                        'describe-instances': [],
                    },
                },
                'arg_data': {
                    '': {
                        'aws': {
                            'query': (
                                'query',
                                'string',
                                'aws',
                                '',
                                None,
                                False,
                                False,
                            ),
                            'output': (
                                'output',
                                'string',
                                'aws',
                                '',
                                None,
                                False,
                                False,
                            ),
                        }
                    },
                    'aws.ec2': {'describe-instances': {}},
                },
            }
        )
        self.parser = parser.CLIParser(index)
        self.completer = basic.QueryCompleter(
            self.cli_fetcher, response_filter=filters.fuzzy_filter
        )

    def _assert_in_completions(self, name, completions):
        self.assertIn(name, [completion.name for completion in completions])

    def _assert_not_in_completions(self, name, completions):
        self.assertNotIn(name, [completion.name for completion in completions])

    def test_get_completion(self):
        parsed = self.parser.parse('aws ec2 describe-instances --query ')
        completions = self.completer.complete(parsed)
        self._assert_in_completions('Reservations', completions)

    def test_complete_on_brackets(self):
        parsed = self.parser.parse(
            'aws ec2 describe-instances --query Reservations[].'
        )
        completions = self.completer.complete(parsed)
        self._assert_in_completions('Reservations[].Groups', completions)

    def test_complete_on_brackets_with_content(self):
        parsed = self.parser.parse(
            'aws ec2 describe-instances --query Reservations[65].'
        )
        completions = self.completer.complete(parsed)
        self._assert_in_completions('Reservations[65].Groups', completions)

    def test_filter_completions(self):
        parsed = self.parser.parse(
            'aws ec2 describe-instances --query Reservations[65].id'
        )
        completions = self.completer.complete(parsed)
        self._assert_in_completions('Reservations[65].OwnerId', completions)
        self._assert_not_in_completions('Reservations[65].Groups', completions)

    def test_not_run_wo_cli_fetcher(self):
        parsed = self.parser.parse(
            'aws ec2 describe-instances --query Reservations[65].id'
        )
        completer = basic.QueryCompleter()
        self.assertIsNone(completer.complete(parsed))

    def test_return_empty_list_when_query_invalid(self):
        parsed = self.parser.parse(
            'aws ec2 describe-instances --query Reservations{65].id'
        )
        completions = self.completer.complete(parsed)
        self.assertEqual([], completions)

    def test_return_empty_list_if_output_is_list_and_expression_is_field(self):
        parsed = self.parser.parse(
            'aws ec2 describe-instances --query Reservations.'
        )
        completions = self.completer.complete(parsed)
        self.assertEqual([], completions)

    def test_return_empty_list_if_output_is_list_and_last_child_is_field(self):
        parsed = self.parser.parse(
            'aws ec2 describe-instances --query Reservations[0].Groups.'
        )
        completions = self.completer.complete(parsed)
        self.assertEqual([], completions)



import errno
import os
import re
import signal
import stat
import subprocess
from configparser import RawConfigParser
from datetime import datetime
from urllib.parse import urlsplit

from dateutil.relativedelta import relativedelta
from dateutil.tz import tzlocal, tzutc

from awscli.compat import urlparse
from awscli.customizations.codeartifact.login import (
    BaseLogin,
    CommandFailedError,
    DotNetLogin,
    NpmLogin,
    NuGetLogin,
    PipLogin,
    SwiftLogin,
    TwineLogin,
    get_relative_expiration_time,
)
from awscli.testutils import FileCreator, mock, skip_if_windows, unittest


class TestBaseLogin(unittest.TestCase):
    def setUp(self):
        self.domain = 'domain'
        self.domain_owner = 'domain-owner'
        self.package_format = 'npm'
        self.repository = 'repository'
        self.auth_token = 'auth-token'
        self.expiration = (
            datetime.now(tzlocal())
            + relativedelta(hours=10)
            + relativedelta(minutes=9)
        ).replace(microsecond=0)
        self.endpoint = (
            f'https://{self.domain}-{self.domain_owner}.codeartifact.aws.'
            f'a2z.com/{self.package_format}/{self.repository}/'
        )

        self.subprocess_utils = mock.Mock()

        self.test_subject = BaseLogin(
            self.auth_token,
            self.expiration,
            self.endpoint,
            self.domain,
            self.repository,
            self.subprocess_utils,
        )

    def test_login(self):
        with self.assertRaises(NotImplementedError):
            self.test_subject.login()

    def test_get_commands(self):
        with self.assertRaises(NotImplementedError):
            self.test_subject.get_commands(self.endpoint, self.auth_token)

    def test_run_commands_command_failed(self):
        error_to_be_caught = subprocess.CalledProcessError(
            returncode=1,
            cmd=['cmd'],
            output=None,
            stderr=b'Command error message.',
        )
        self.subprocess_utils.run.side_effect = error_to_be_caught
        with self.assertRaisesRegex(
            CommandFailedError,
            rf"{re.escape(str(error_to_be_caught))}"
            rf" Stderr from command:\nCommand error message.",
        ):
            self.test_subject._run_commands('tool', ['cmd'])

    def test_run_commands_command_failed_redact_auth_token(self):
        error_to_be_caught = subprocess.CalledProcessError(
            returncode=1,
            cmd=['cmd', 'with', 'auth-token', 'present'],
            output=None,
            stderr=b'Command error message.',
        )
        self.subprocess_utils.run.side_effect = error_to_be_caught
        with self.assertRaisesRegex(
            CommandFailedError,
            (
                r"(?=.*cmd)(?=.*with)(?!.*auth-token)(?=.*present)"
                r"(?=.*Stderr from command:\nCommand error message.)"
            ),
        ):
            self.test_subject._run_commands('tool', ['cmd'])

    def test_run_commands_nonexistent_command(self):
        self.subprocess_utils.run.side_effect = OSError(
            errno.ENOENT, 'not found error'
        )
        tool = 'NotSupported'
        with self.assertRaisesRegex(ValueError, f'{tool} was not found.'):
            self.test_subject._run_commands(tool, ['echo', tool])

    def test_run_commands_unhandled_error(self):
        self.subprocess_utils.run.side_effect = OSError(
            errno.ENOSYS, 'unhandled error'
        )
        tool = 'NotSupported'
        with self.assertRaisesRegex(OSError, 'unhandled error'):
            self.test_subject._run_commands(tool, ['echo', tool])


def handle_timeout(signum, frame, test_name):
    raise TimeoutError(f"{test_name} timed out!!")


class TestSwiftLogin(unittest.TestCase):
    def setUp(self):
        self.domain = 'domain'
        self.domain_owner = 'domain-owner'
        self.package_format = 'swift'
        self.repository = 'repository'
        self.auth_token = 'auth-token'
        self.namespace = 'namespace'
        self.expiration = (
            datetime.now(tzlocal())
            + relativedelta(hours=10)
            + relativedelta(minutes=9)
        ).replace(microsecond=0)
        self.endpoint = (
            f'https://{self.domain}-{self.domain_owner}.codeartifact'
            f'.aws.a2z.com/{self.package_format}/{self.repository}/'
        )
        self.hostname = urlparse.urlparse(self.endpoint).hostname
        self.new_entry = SwiftLogin.DEFAULT_NETRC_FMT.format(
            hostname=self.hostname, auth_token=self.auth_token
        )

        self.file_creator = FileCreator()
        self.test_netrc_path = self.file_creator.full_path('netrc')
        self.get_netrc_path_patch = mock.patch(
            'awscli.customizations.codeartifact.login.SwiftLogin'
            '.get_netrc_path'
        )
        self.get_netrc_path_mock = self.get_netrc_path_patch.start()
        self.get_netrc_path_mock.return_value = self.test_netrc_path

        self.base_command = ['swift', 'package-registry', 'set', self.endpoint]
        self.macos_commands = [
            self.base_command[:],
            [
                'swift',
                'package-registry',
                'login',
                self.endpoint + 'login',
                '--token',
                self.auth_token,
            ],
        ]
        self.non_macos_commands = [
            self.base_command[:],
            ['swift', 'package-registry', 'login', self.endpoint + 'login'],
        ]

        self.subprocess_utils = mock.Mock()

        self.test_subject = SwiftLogin(
            self.auth_token,
            self.expiration,
            self.endpoint,
            self.domain,
            self.repository,
            self.subprocess_utils,
        )

    def tearDown(self):
        self.get_netrc_path_patch.stop()
        self.file_creator.remove_all()

    def _assert_netrc_has_expected_content(self, expected_contents):
        with open(self.test_netrc_path) as file:
            actual_contents = file.read()
            self.assertEqual(expected_contents, actual_contents)

    def _assert_netrc_has_expected_permissions(self):
        file_stat = os.stat(self.test_netrc_path)
        file_mode = file_stat.st_mode
        self.assertTrue(stat.S_IRUSR & file_mode)
        self.assertTrue(stat.S_IWUSR & file_mode)

    def test_get_netrc_path(self):
        self.assertEqual(SwiftLogin.get_netrc_path(), self.test_netrc_path)

    def test_regex_only_match_escaped_hostname(self):
        pattern = re.compile(
            SwiftLogin.NETRC_REGEX_FMT.format(
                escaped_hostname=re.escape(self.hostname)
            ),
            re.M,
        )
        bad_endpoint = (
            f'https://{self.domain}-{self.domain_owner}-codeartifact'
            f'-aws-a2z-com/{self.package_format}/{self.repository}/'
        )
        bad_hostname = urlparse.urlparse(bad_endpoint).hostname
        bad_entry = SwiftLogin.DEFAULT_NETRC_FMT.format(
            hostname=bad_hostname, auth_token=self.auth_token
        )
        self.assertTrue(pattern.match(self.new_entry))
        self.assertFalse(pattern.match(bad_entry))

    def test_create_netrc_if_not_exist(self):
        self.assertFalse(os.path.exists(self.test_netrc_path))
        self.test_subject._update_netrc_entry(
            self.hostname, 'a new entry', self.test_netrc_path
        )
        self.assertTrue(os.path.exists(self.test_netrc_path))
        self._assert_netrc_has_expected_permissions()
        self._assert_netrc_has_expected_content('a new entry\n')

    def test_replacement_token_has_backslash(self):
        existing_content = f'machine {self.hostname} login token password expired-auth-token\n'
        with open(self.test_netrc_path, 'w+') as f:
            f.write(existing_content)
        self.test_subject.auth_token = r'new-token_.\1\g<entry_start>\n\w'
        # make sure it uses re.sub() to replace the token
        self.test_subject._update_netrc_entry(
            self.hostname, '', self.test_netrc_path
        )
        self.assertTrue(os.path.exists(self.test_netrc_path))
        self._assert_netrc_has_expected_content(
            f'machine {self.hostname} login token password {self.test_subject.auth_token}\n'
        )

    def test_update_netrc_with_existing_entry(self):
        existing_content = f'machine {self.hostname} login token password expired-auth-token\n'

        expected_content = f'{self.new_entry}\n'
        with open(self.test_netrc_path, 'w+') as f:
            f.write(existing_content)

        self.test_subject._update_netrc_entry(
            self.hostname, self.new_entry, self.test_netrc_path
        )
        self._assert_netrc_has_expected_content(expected_content)

    def test_update_netrc_with_existing_entry_in_between(self):
        existing_content = (
            f'some random line above\n'
            f'machine {self.hostname} login token password expired-auth-token\n'
            f'some random line below\n'
        )

        expected_content = (
            f'some random line above\n'
            f'{self.new_entry}\n'
            f'some random line below\n'
        )
        with open(self.test_netrc_path, 'w+') as f:
            f.write(existing_content)

        self.test_subject._update_netrc_entry(
            self.hostname, self.new_entry, self.test_netrc_path
        )
        self._assert_netrc_has_expected_content(expected_content)

    def test_append_netrc_without_ending_newline(self):
        existing_content = 'machine host login user password 1234'

        expected_content = (
            f'machine host login user password 1234\n' f'{self.new_entry}\n'
        )
        with open(self.test_netrc_path, 'w+') as f:
            f.write(existing_content)

        self.test_subject._update_netrc_entry(
            self.hostname, self.new_entry, self.test_netrc_path
        )
        self._assert_netrc_has_expected_content(expected_content)

    def test_append_netrc_with_ending_newline(self):
        existing_content = 'machine host login user password 1234\n'

        expected_content = (
            f'machine host login user password 1234\n' f'{self.new_entry}\n'
        )
        with open(self.test_netrc_path, 'w+') as f:
            f.write(existing_content)

        self.test_subject._update_netrc_entry(
            self.hostname, self.new_entry, self.test_netrc_path
        )
        self._assert_netrc_has_expected_content(expected_content)

    def test_update_netrc_with_multiple_spaces_and_newlines(self):
        existing_content = (
            f' machine   {self.hostname}\n'
            f'   login token \n'
            f'password expired-auth-token  \n'
            f'\n'
            f'machine example1.com\n'
            f' login user1 \n'
            f'password  1234\n'
        )
        expected_content = (
            f' machine   {self.hostname}\n'
            f'   login token \n'
            f'password auth-token  \n'
            f'\n'
            f'machine example1.com\n'
            f' login user1 \n'
            f'password  1234\n'
        )
        with open(self.test_netrc_path, 'w+') as f:
            f.write(existing_content)

        self.test_subject._update_netrc_entry(
            self.hostname, self.new_entry, self.test_netrc_path
        )
        self._assert_netrc_has_expected_content(expected_content)

    def test_update_netrc_with_multiple_existing_entries(self):
        existing_content = (
            f'machine {self.hostname} login token password expired-auth-token-1\n'
            f'machine {self.hostname} login token password expired-auth-token-2\n'
        )
        expected_content = f'{self.new_entry}\n' f'{self.new_entry}\n'
        with open(self.test_netrc_path, 'w+') as f:
            f.write(existing_content)

        self.test_subject._update_netrc_entry(
            self.hostname, self.new_entry, self.test_netrc_path
        )
        self._assert_netrc_has_expected_content(expected_content)

    @mock.patch('awscli.customizations.codeartifact.login.is_macos', True)
    def test_login_macos(self):
        self.test_subject.login()
        expected_calls = [
            mock.call(command, capture_output=True, check=True)
            for command in self.macos_commands
        ]
        self.subprocess_utils.run.assert_has_calls(
            expected_calls, any_order=True
        )

    @mock.patch('awscli.customizations.codeartifact.login.is_macos', False)
    def test_login_non_macos(self):
        self.test_subject.login()
        expected_calls = [
            mock.call(command, capture_output=True, check=True)
            for command in self.non_macos_commands
        ]
        self.subprocess_utils.run.assert_has_calls(
            expected_calls, any_order=True
        )

    def test_login_swift_tooling_error(self):
        self.subprocess_utils.run.side_effect = subprocess.CalledProcessError(
            returncode=1, cmd='swift command', stderr=b''
        )
        with self.assertRaises(CommandFailedError):
            self.test_subject.login()

    def test_login_swift_not_installed(self):
        self.subprocess_utils.run.side_effect = OSError(
            errno.ENOENT, 'not found error'
        )
        with self.assertRaisesRegex(
            ValueError, 'swift was not found. Please verify installation.'
        ):
            self.test_subject.login()

    def test_get_scope(self):
        expected_value = 'namespace'
        scope = self.test_subject.get_scope(self.namespace)
        self.assertEqual(scope, expected_value)

    def test_get_scope_none_namespace(self):
        expected_value = None
        scope = self.test_subject.get_scope(None)
        self.assertEqual(scope, expected_value)

    def test_get_scope_invalid_leading_character(self):
        with self.assertRaises(ValueError):
            self.test_subject.get_scope(f'.{self.namespace}')

    def test_get_scope_invalid_length(self):
        with self.assertRaises(ValueError):
            self.test_subject.get_scope("a" * 40)

    @mock.patch('awscli.customizations.codeartifact.login.is_macos', True)
    def test_get_commands_macos(self):
        commands = self.test_subject.get_commands(
            self.endpoint, self.auth_token
        )
        self.assertCountEqual(commands, self.macos_commands)

    @mock.patch('awscli.customizations.codeartifact.login.is_macos', False)
    def test_get_commands_non_macos(self):
        commands = self.test_subject.get_commands(
            self.endpoint, self.auth_token
        )
        self.assertCountEqual(commands, self.non_macos_commands)

    @mock.patch('awscli.customizations.codeartifact.login.is_macos', True)
    def test_get_commands_with_scope_macos(self):
        commands = self.test_subject.get_commands(
            self.endpoint, self.auth_token, scope=self.namespace
        )
        self.macos_commands[0] += ['--scope', self.namespace]
        self.assertCountEqual(commands, self.macos_commands)

    @mock.patch('awscli.customizations.codeartifact.login.is_macos', False)
    def test_get_commands_with_scope_non_macos(self):
        commands = self.test_subject.get_commands(
            self.endpoint, self.auth_token, scope=self.namespace
        )
        self.non_macos_commands[0] += ['--scope', self.namespace]
        self.assertCountEqual(commands, self.non_macos_commands)

    def test_login_dry_run(self):
        self.test_subject.login(dry_run=True)
        self.subprocess_utils.check_output.assert_not_called()
        self.assertFalse(os.path.exists(self.test_netrc_path))

    @skip_if_windows("Unix file permissions are not supported on Windows.")
    @mock.patch('awscli.customizations.codeartifact.login.is_macos', False)
    def test_login_adjusts_permissions_on_preexisting_file(self):
        existing_content = (
            f'machine {self.hostname} login token password old-token\n'
        )
        with open(self.test_netrc_path, 'w') as f:
            f.write(existing_content)
        os.chmod(self.test_netrc_path, 0o644)
        self.test_subject.login()
        file_mode = os.stat(self.test_netrc_path).st_mode
        self.assertEqual(stat.S_IMODE(file_mode), 0o600)

    @skip_if_windows("Unix file permissions are not supported on Windows.")
    @mock.patch('awscli.customizations.codeartifact.login.is_macos', False)
    @mock.patch('sys.stderr')
    @mock.patch('os.chmod', side_effect=OSError(
        errno.EPERM, 'Operation not permitted'
    ))
    def test_login_writes_error_when_chmod_fails(
        self, mock_chmod, mock_stderr
    ):
        existing_content = (
            f'machine {self.hostname} login token password old-token\n'
        )
        with open(self.test_netrc_path, 'w') as f:
            f.write(existing_content)
        self.test_subject.login()
        mock_stderr.write.assert_any_call(
            'Unable to set file permissions for %s: '
            '[Errno 1] Operation not permitted%s'
            % (self.test_netrc_path, os.linesep)
        )

class TestNuGetLogin(unittest.TestCase):
    _NUGET_INDEX_URL_FMT = NuGetLogin._NUGET_INDEX_URL_FMT
    _NUGET_SOURCES_LIST_RESPONSE = b"""\
Registered Sources:
  1. Source Name 1 [Enabled]
     https://source1.com/index.json
  2. Ab[.d7  $#!],   [Disabled]
     https://source2.com/index.json"""

    _NUGET_SOURCES_LIST_RESPONSE_BACKTRACKING = b'1.' + b' ' * 10000 + b'a'

    _NUGET_SOURCES_LIST_RESPONSE_WITH_SPACE = b"""\
                Registered Sources:

                  1. Source Name 1 [Enabled]
                     https://source1.com/index.json
                  2. Ab[.d7  $#!],   [Disabled]
                     https://source2.com/index.json"""

    def setUp(self):
        self.domain = 'domain'
        self.domain_owner = 'domain-owner'
        self.package_format = 'nuget'
        self.repository = 'repository'
        self.auth_token = 'auth-token'
        self.expiration = (
            datetime.now(tzlocal())
            + relativedelta(hours=10)
            + relativedelta(minutes=9)
        ).replace(microsecond=0)
        self.endpoint = (
            f'https://{self.domain}-{self.domain_owner}.codeartifact.aws.'
            f'a2z.com/{self.package_format}/{self.repository}/'
        )

        self.nuget_index_url = self._NUGET_INDEX_URL_FMT.format(
            endpoint=self.endpoint,
        )
        self.source_name = self.domain + '/' + self.repository
        self.list_operation_command = [
            'nuget',
            'sources',
            'list',
            '-format',
            'detailed',
        ]
        self.add_operation_command = [
            'nuget',
            'sources',
            'add',
            '-name',
            self.source_name,
            '-source',
            self.nuget_index_url,
            '-username',
            'aws',
            '-password',
            self.auth_token,
        ]
        self.update_operation_command = [
            'nuget',
            'sources',
            'update',
            '-name',
            self.source_name,
            '-source',
            self.nuget_index_url,
            '-username',
            'aws',
            '-password',
            self.auth_token,
        ]

        self.subprocess_utils = mock.Mock()

        self.test_subject = NuGetLogin(
            self.auth_token,
            self.expiration,
            self.endpoint,
            self.domain,
            self.repository,
            self.subprocess_utils,
        )

    def test_login(self):
        self.subprocess_utils.check_output.return_value = (
            self._NUGET_SOURCES_LIST_RESPONSE
        )
        self.test_subject.login()
        self.subprocess_utils.check_output.assert_any_call(
            self.list_operation_command, stderr=self.subprocess_utils.PIPE
        )
        self.subprocess_utils.run.assert_called_with(
            self.add_operation_command, capture_output=True, check=True
        )

    def test_login_dry_run(self):
        self.subprocess_utils.check_output.return_value = (
            self._NUGET_SOURCES_LIST_RESPONSE
        )
        self.test_subject.login(dry_run=True)
        self.subprocess_utils.check_output.assert_called_once_with(
            ['nuget', 'sources', 'list', '-format', 'detailed'],
            stderr=self.subprocess_utils.PIPE,
        )

    def test_login_old_nuget(self):
        self.subprocess_utils.check_output.return_value = (
            self._NUGET_SOURCES_LIST_RESPONSE_WITH_SPACE
        )
        self.test_subject.login()
        self.subprocess_utils.check_output.assert_any_call(
            self.list_operation_command, stderr=self.subprocess_utils.PIPE
        )
        self.subprocess_utils.run.assert_called_with(
            self.add_operation_command, capture_output=True, check=True
        )

    def test_login_dry_run_old_nuget(self):
        self.subprocess_utils.check_output.return_value = (
            self._NUGET_SOURCES_LIST_RESPONSE_WITH_SPACE
        )
        self.test_subject.login(dry_run=True)
        self.subprocess_utils.check_output.assert_called_once_with(
            ['nuget', 'sources', 'list', '-format', 'detailed'],
            stderr=self.subprocess_utils.PIPE,
        )

    def test_login_source_name_already_exists(self):
        list_response = (
            'Registered Sources:\n'
            '  1.  ' + self.source_name + ' [ENABLED]\n'
            '      https://source.com/index.json'
        )
        self.subprocess_utils.check_output.return_value = list_response.encode(
            'utf-8'
        )
        self.test_subject.login()
        self.subprocess_utils.run.assert_called_with(
            self.update_operation_command, capture_output=True, check=True
        )

    def test_login_source_url_already_exists_old_nuget(self):
        non_default_source_name = 'Source Name'
        list_response = (
            'Registered Sources:\n'
            '\n'
            '  1. ' + non_default_source_name + ' [ENABLED]\n'
            '      ' + self.nuget_index_url
        )
        self.subprocess_utils.check_output.return_value = list_response.encode(
            'utf-8'
        )
        self.test_subject.login()
        self.subprocess_utils.run.assert_called_with(
            [
                'nuget',
                'sources',
                'update',
                '-name',
                non_default_source_name,
                '-source',
                self.nuget_index_url,
                '-username',
                'aws',
                '-password',
                self.auth_token,
            ],
            capture_output=True,
            check=True,
        )

    def test_login_source_url_already_exists(self):
        non_default_source_name = 'Source Name'
        list_response = (
            'Registered Sources:\n'
            '  1. ' + non_default_source_name + ' [ENABLED]\n'
            '      ' + self.nuget_index_url
        )
        self.subprocess_utils.check_output.return_value = list_response.encode(
            'utf-8'
        )
        self.test_subject.login()
        self.subprocess_utils.run.assert_called_with(
            [
                'nuget',
                'sources',
                'update',
                '-name',
                non_default_source_name,
                '-source',
                self.nuget_index_url,
                '-username',
                'aws',
                '-password',
                self.auth_token,
            ],
            capture_output=True,
            check=True,
        )

    def test_login_nuget_not_installed(self):
        self.subprocess_utils.check_output.side_effect = OSError(
            errno.ENOENT, 'not found error'
        )
        with self.assertRaisesRegex(
            ValueError, 'nuget was not found. Please verify installation.'
        ):
            self.test_subject.login()

    @skip_if_windows("Windows does not support signal.SIGALRM.")
    def test_login_nuget_sources_listed_with_backtracking(self):
        self.subprocess_utils.check_output.return_value = (
            self._NUGET_SOURCES_LIST_RESPONSE_BACKTRACKING
        )
        signal.signal(
            signal.SIGALRM,
            lambda signum, frame: handle_timeout(signum, frame, self.id()),
        )
        signal.alarm(10)
        self.test_subject.login()
        signal.alarm(0)
        self.subprocess_utils.check_output.assert_any_call(
            self.list_operation_command, stderr=self.subprocess_utils.PIPE
        )


class TestDotNetLogin(unittest.TestCase):
    _NUGET_INDEX_URL_FMT = NuGetLogin._NUGET_INDEX_URL_FMT
    _NUGET_SOURCES_LIST_RESPONSE = b"""\
Registered Sources:
  1. Source Name 1 [Enabled]
     https://source1.com/index.json
  2. Ab[.d7  $#!],   [Disabled]
     https://source2.com/index.json"""

    _NUGET_SOURCES_LIST_RESPONSE_BACKTRACKING = b'1.' + b' ' * 10000 + b'a'

    _NUGET_SOURCES_LIST_RESPONSE_WITH_EXTRA_NON_LIST_TEXT = b"""\
Welcome to dotnet 2.0!

Registered Sources:
  1. Source Name 1 [Enabled]
     https://source1.com/index.json
  2. ati-nugetserver [Disabled]
     http://atinugetserver-env.elasticbeanstalk.com/nuget
warn : You are running the 'list source' operation with an 'HTTP' source,
'ati-nugetserver' [http://atinugetserver-env..elasticbeanstalk.com/nuget]'.
Non-HTTPS access will be removed in a future version. Consider migrating
to an 'HTTPS' source."""

    def setUp(self):
        self.domain = 'domain'
        self.domain_owner = 'domain-owner'
        self.package_format = 'nuget'
        self.repository = 'repository'
        self.auth_token = 'auth-token'
        self.expiration = (
            datetime.now(tzlocal())
            + relativedelta(hours=10)
            + relativedelta(minutes=9)
        ).replace(microsecond=0)
        self.endpoint = (
            f'https://{self.domain}-{self.domain_owner}.codeartifact.aws.'
            f'a2z.com/{self.package_format}/{self.repository}/'
        )

        self.nuget_index_url = self._NUGET_INDEX_URL_FMT.format(
            endpoint=self.endpoint,
        )
        self.source_name = self.domain + '/' + self.repository
        self.list_operation_command = [
            'dotnet',
            'nuget',
            'list',
            'source',
            '--format',
            'detailed',
        ]
        self.add_operation_command_windows = [
            'dotnet',
            'nuget',
            'add',
            'source',
            self.nuget_index_url,
            '--name',
            self.source_name,
            '--username',
            'aws',
            '--password',
            self.auth_token,
        ]
        self.add_operation_command_non_windows = [
            'dotnet',
            'nuget',
            'add',
            'source',
            self.nuget_index_url,
            '--name',
            self.source_name,
            '--username',
            'aws',
            '--password',
            self.auth_token,
            '--store-password-in-clear-text',
        ]
        self.update_operation_command_windows = [
            'dotnet',
            'nuget',
            'update',
            'source',
            self.source_name,
            '--source',
            self.nuget_index_url,
            '--username',
            'aws',
            '--password',
            self.auth_token,
        ]
        self.update_operation_command_non_windows = [
            'dotnet',
            'nuget',
            'update',
            'source',
            self.source_name,
            '--source',
            self.nuget_index_url,
            '--username',
            'aws',
            '--password',
            self.auth_token,
            '--store-password-in-clear-text',
        ]

        self.subprocess_utils = mock.Mock()

        self.test_subject = DotNetLogin(
            self.auth_token,
            self.expiration,
            self.endpoint,
            self.domain,
            self.repository,
            self.subprocess_utils,
        )

    @mock.patch('awscli.customizations.codeartifact.login.is_windows', False)
    def test_login(self):
        self.subprocess_utils.check_output.return_value = (
            self._NUGET_SOURCES_LIST_RESPONSE
        )
        self.test_subject.login()
        self.subprocess_utils.check_output.assert_any_call(
            self.list_operation_command, stderr=self.subprocess_utils.PIPE
        )
        self.subprocess_utils.run.assert_called_with(
            self.add_operation_command_non_windows,
            capture_output=True,
            check=True,
        )

    @mock.patch('awscli.customizations.codeartifact.login.is_windows', True)
    def test_login_on_windows(self):
        self.subprocess_utils.check_output.return_value = (
            self._NUGET_SOURCES_LIST_RESPONSE
        )
        self.test_subject.login()
        self.subprocess_utils.check_output.assert_any_call(
            self.list_operation_command, stderr=self.subprocess_utils.PIPE
        )
        self.subprocess_utils.run.assert_called_with(
            self.add_operation_command_windows, capture_output=True, check=True
        )

    def test_login_dry_run(self):
        self.subprocess_utils.check_output.return_value = (
            self._NUGET_SOURCES_LIST_RESPONSE
        )
        self.test_subject.login(dry_run=True)
        self.subprocess_utils.check_output.assert_called_once_with(
            self.list_operation_command, stderr=self.subprocess_utils.PIPE
        )

    @mock.patch('awscli.customizations.codeartifact.login.is_windows', False)
    def test_login_sources_listed_with_extra_non_list_text(self):
        self.subprocess_utils.check_output.return_value = (
            self._NUGET_SOURCES_LIST_RESPONSE_WITH_EXTRA_NON_LIST_TEXT
        )
        self.test_subject.login()
        self.subprocess_utils.check_output.assert_any_call(
            self.list_operation_command, stderr=self.subprocess_utils.PIPE
        )
        self.subprocess_utils.run.assert_called_with(
            self.add_operation_command_non_windows,
            capture_output=True,
            check=True,
        )

    @mock.patch('awscli.customizations.codeartifact.login.is_windows', True)
    def test_login_sources_listed_with_extra_non_list_text_on_windows(self):
        self.subprocess_utils.check_output.return_value = (
            self._NUGET_SOURCES_LIST_RESPONSE_WITH_EXTRA_NON_LIST_TEXT
        )
        self.test_subject.login()
        self.subprocess_utils.check_output.assert_any_call(
            self.list_operation_command, stderr=self.subprocess_utils.PIPE
        )
        self.subprocess_utils.run.assert_called_with(
            self.add_operation_command_windows, capture_output=True, check=True
        )

    def test_login_sources_listed_with_extra_non_list_text_dry_run(self):
        self.subprocess_utils.check_output.return_value = (
            self._NUGET_SOURCES_LIST_RESPONSE_WITH_EXTRA_NON_LIST_TEXT
        )
        self.test_subject.login(dry_run=True)
        self.subprocess_utils.check_output.assert_called_once_with(
            self.list_operation_command, stderr=self.subprocess_utils.PIPE
        )

    @skip_if_windows("Windows does not support signal.SIGALRM.")
    @mock.patch('awscli.customizations.codeartifact.login.is_windows', False)
    def test_login_dotnet_sources_listed_with_backtracking(self):
        self.subprocess_utils.check_output.return_value = (
            self._NUGET_SOURCES_LIST_RESPONSE_BACKTRACKING
        )
        signal.signal(
            signal.SIGALRM,
            lambda signum, frame: handle_timeout(signum, frame, self.id()),
        )
        signal.alarm(10)
        self.test_subject.login()
        signal.alarm(0)

    @skip_if_windows("Windows does not support signal.SIGALRM.")
    @mock.patch('awscli.customizations.codeartifact.login.is_windows', True)
    def test_login_dotnet_sources_listed_with_backtracking_windows(self):
        self.subprocess_utils.check_output.return_value = (
            self._NUGET_SOURCES_LIST_RESPONSE_BACKTRACKING
        )
        signal.signal(
            signal.SIGALRM,
            lambda signum, frame: handle_timeout(signum, frame, self.id()),
        )
        signal.alarm(10)
        self.test_subject.login()
        signal.alarm(0)

    @mock.patch('awscli.customizations.codeartifact.login.is_windows', False)
    def test_login_source_name_already_exists(self):
        list_response = (
            'Registered Sources:\n'
            '  1.  ' + self.source_name + ' [ENABLED]\n'
            '      https://source.com/index.json'
        )
        self.subprocess_utils.check_output.return_value = list_response.encode(
            'utf-8'
        )
        self.test_subject.login()
        self.subprocess_utils.run.assert_called_with(
            self.update_operation_command_non_windows,
            capture_output=True,
            check=True,
        )

    @mock.patch('awscli.customizations.codeartifact.login.is_windows', True)
    def test_login_source_name_already_exists_on_windows(self):
        list_response = (
            'Registered Sources:\n'
            '  1.  ' + self.source_name + ' [ENABLED]\n'
            '      https://source.com/index.json'
        )
        self.subprocess_utils.check_output.return_value = list_response.encode(
            'utf-8'
        )
        self.test_subject.login()
        self.subprocess_utils.run.assert_called_with(
            self.update_operation_command_windows,
            capture_output=True,
            check=True,
        )

    @mock.patch('awscli.customizations.codeartifact.login.is_windows', True)
    def test_login_source_url_already_exists(self):
        non_default_source_name = 'Source Name'
        list_response = (
            'Registered Sources:\n'
            '  1. ' + non_default_source_name + ' [ENABLED]\n'
            '      ' + self.nuget_index_url
        )
        self.subprocess_utils.check_output.return_value = list_response.encode(
            'utf-8'
        )
        self.test_subject.login()
        self.subprocess_utils.run.assert_called_with(
            [
                'dotnet',
                'nuget',
                'update',
                'source',
                non_default_source_name,
                '--source',
                self.nuget_index_url,
                '--username',
                'aws',
                '--password',
                self.auth_token,
            ],
            capture_output=True,
            check=True,
        )

    def test_login_dotnet_not_installed(self):
        self.subprocess_utils.check_output.side_effect = OSError(
            errno.ENOENT, 'not found error'
        )
        with self.assertRaisesRegex(
            ValueError, 'dotnet was not found. Please verify installation.'
        ):
            self.test_subject.login()


class TestNpmLogin(unittest.TestCase):
    NPM_CMD = NpmLogin.NPM_CMD

    def setUp(self):
        self.domain = 'domain'
        self.domain_owner = 'domain-owner'
        self.package_format = 'npm'
        self.repository = 'repository'
        self.auth_token = 'auth-token'
        self.namespace = 'namespace'
        self.expiration = (
            datetime.now(tzlocal())
            + relativedelta(hours=10)
            + relativedelta(minutes=9)
        ).replace(microsecond=0)
        self.endpoint = (
            f'https://{self.domain}-{self.domain_owner}.codeartifact.aws.'
            f'a2z.com/{self.package_format}/{self.repository}/'
        )

        repo_uri = urlsplit(self.endpoint)
        always_auth_config = f'//{repo_uri.netloc}{repo_uri.path}:always-auth'
        auth_token_config = f'//{repo_uri.netloc}{repo_uri.path}:_authToken'
        self.commands = []
        self.commands.append(
            [self.NPM_CMD, 'config', 'set', 'registry', self.endpoint]
        )
        self.commands.append(
            [self.NPM_CMD, 'config', 'set', always_auth_config, 'true']
        )
        self.commands.append(
            [self.NPM_CMD, 'config', 'set', auth_token_config, self.auth_token]
        )

        self.subprocess_utils = mock.Mock()

        self.test_subject = NpmLogin(
            self.auth_token,
            self.expiration,
            self.endpoint,
            self.domain,
            self.repository,
            self.subprocess_utils,
        )

    def test_login(self):
        self.test_subject.login()
        expected_calls = [
            mock.call(command, capture_output=True, check=True)
            for command in self.commands
        ]
        self.subprocess_utils.run.assert_has_calls(
            expected_calls, any_order=True
        )

    def test_login_always_auth_error_ignored(self):
        """Test login ignores error for always-auth.

        This test is for NPM version >= 9 where the support of 'always-auth'
        has been dropped. Running the command to set config gives a non-zero
        exit code. This is to make sure that login ignores that error and all
        other commands executes successfully.
        """

        def side_effect(command, capture_output, check):
            """Set side_effect for always-auth config setting command"""
            if any('always-auth' in arg for arg in command):
                raise subprocess.CalledProcessError(returncode=1, cmd=command)

            return mock.DEFAULT

        self.subprocess_utils.run.side_effect = side_effect
        expected_calls = []

        for command in self.commands:
            expected_calls.append(
                mock.call(command, capture_output=True, check=True)
            )
        self.test_subject.login()

        self.subprocess_utils.run.assert_has_calls(
            expected_calls, any_order=True
        )

    def test_get_scope(self):
        expected_value = f'@{self.namespace}'
        scope = self.test_subject.get_scope(self.namespace)
        self.assertEqual(scope, expected_value)

    def test_get_scope_none_namespace(self):
        expected_value = None
        scope = self.test_subject.get_scope(None)
        self.assertEqual(scope, expected_value)

    def test_get_scope_invalid_name(self):
        with self.assertRaises(ValueError):
            self.test_subject.get_scope(f'.{self.namespace}')

    def test_get_scope_without_prefix(self):
        expected_value = f'@{self.namespace}'
        scope = self.test_subject.get_scope(f'@{self.namespace}')
        self.assertEqual(scope, expected_value)

    def test_get_commands(self):
        commands = self.test_subject.get_commands(
            self.endpoint, self.auth_token
        )
        self.assertCountEqual(commands, self.commands)

    def test_get_commands_with_scope(self):
        commands = self.test_subject.get_commands(
            self.endpoint, self.auth_token, scope=self.namespace
        )
        self.commands[0][3] = f'{self.namespace}:registry'
        self.assertCountEqual(commands, self.commands)

    def test_login_dry_run(self):
        self.test_subject.login(dry_run=True)
        self.subprocess_utils.assert_not_called()


class TestPipLogin(unittest.TestCase):
    PIP_INDEX_URL_FMT = PipLogin.PIP_INDEX_URL_FMT

    def setUp(self):
        self.domain = 'domain'
        self.domain_owner = 'domain-owner'
        self.package_format = 'pip'
        self.repository = 'repository'
        self.auth_token = 'auth-token'
        self.expiration = (
            datetime.now(tzlocal())
            + relativedelta(years=1)
            + relativedelta(months=9)
        ).replace(microsecond=0)
        self.endpoint = (
            f'https://{self.domain}-{self.domain_owner}.codeartifact.aws.'
            f'a2z.com/{self.package_format}/{self.repository}/'
        )

        repo_uri = urlsplit(self.endpoint)
        self.pip_index_url = self.PIP_INDEX_URL_FMT.format(
            scheme=repo_uri.scheme,
            auth_token=self.auth_token,
            netloc=repo_uri.netloc,
            path=repo_uri.path,
        )

        self.subprocess_utils = mock.Mock()

        self.test_subject = PipLogin(
            self.auth_token,
            self.expiration,
            self.endpoint,
            self.domain,
            self.repository,
            self.subprocess_utils,
        )

    def test_get_commands(self):
        expected_commands = [
            ['pip', 'config', 'set', 'global.index-url', self.pip_index_url]
        ]
        commands = self.test_subject.get_commands(
            self.endpoint, self.auth_token
        )
        self.assertCountEqual(commands, expected_commands)

    def test_login(self):
        self.test_subject.login()
        self.subprocess_utils.run.assert_called_once_with(
            ['pip', 'config', 'set', 'global.index-url', self.pip_index_url],
            capture_output=True,
            check=True,
        )

    def test_login_dry_run(self):
        self.test_subject.login(dry_run=True)
        self.subprocess_utils.run.assert_not_called()


class TestTwineLogin(unittest.TestCase):
    DEFAULT_PYPI_RC_FMT = TwineLogin.DEFAULT_PYPI_RC_FMT

    def setUp(self):
        self.file_creator = FileCreator()
        self.domain = 'domain'
        self.domain_owner = 'domain-owner'
        self.package_format = 'pip'
        self.repository = 'repository'
        self.auth_token = 'auth-token'
        self.expiration = (
            datetime.now(tzlocal())
            + relativedelta(years=1)
            + relativedelta(months=9)
        ).replace(microsecond=0)
        self.endpoint = (
            f'https://{self.domain}-{self.domain_owner}.codeartifact.aws.'
            f'a2z.com/{self.package_format}/{self.repository}/'
        )
        self.default_pypi_rc = self.DEFAULT_PYPI_RC_FMT.format(
            repository_endpoint=self.endpoint, auth_token=self.auth_token
        )
        self.subprocess_utils = mock.Mock()
        self.test_pypi_rc_path = self.file_creator.full_path('pypirc')
        if not os.path.isdir(os.path.dirname(self.test_pypi_rc_path)):
            os.makedirs(os.path.dirname(self.test_pypi_rc_path))

        self.test_subject = TwineLogin(
            self.auth_token,
            self.expiration,
            self.endpoint,
            self.domain,
            self.repository,
            self.subprocess_utils,
            self.test_pypi_rc_path,
        )

    def tearDown(self):
        self.file_creator.remove_all()

    def _assert_pypi_rc_has_expected_content(
        self, pypi_rc_str, server, repo_url=None, username=None, password=None
    ):
        pypi_rc = RawConfigParser()
        pypi_rc.read_string(pypi_rc_str)

        self.assertIn('distutils', pypi_rc.sections())
        self.assertIn('index-servers', pypi_rc.options('distutils'))
        index_servers = pypi_rc.get('distutils', 'index-servers')
        index_servers = [
            index_server.strip()
            for index_server in index_servers.split('\n')
            if index_server.strip() != ''
        ]
        self.assertIn(server, index_servers)

        if repo_url or username or password:
            self.assertIn(server, pypi_rc.sections())

        if repo_url:
            self.assertIn('repository', pypi_rc.options(server))
            self.assertEqual(pypi_rc.get(server, 'repository'), repo_url)

        if username:
            self.assertIn('username', pypi_rc.options(server))
            self.assertEqual(pypi_rc.get(server, 'username'), username)

        if password:
            self.assertIn('password', pypi_rc.options(server))
            self.assertEqual(pypi_rc.get(server, 'password'), password)

    def test_get_pypi_rc_path(self):
        self.assertEqual(
            TwineLogin.get_pypi_rc_path(),
            os.path.join(os.path.expanduser("~"), ".pypirc"),
        )

    def test_login_pypi_rc_not_found_defaults_set(self):
        self.test_subject.login()

        with open(self.test_pypi_rc_path) as f:
            test_pypi_rc_str = f.read()

        self._assert_pypi_rc_has_expected_content(
            pypi_rc_str=test_pypi_rc_str,
            server='codeartifact',
            repo_url=self.endpoint,
            username='aws',
            password=self.auth_token,
        )

    def test_login_dry_run(self):
        self.test_subject.login(dry_run=True)
        self.subprocess_utils.run.assert_not_called()
        self.assertFalse(os.path.exists(self.test_pypi_rc_path))

    def test_login_existing_pypi_rc_not_clobbered(self):
        existing_pypi_rc = '''\
[distutils]
index-servers=
    pypi
    test

[pypi]
repository: http://www.python.org/pypi/
username: monty
password: JgCXIr5xGG

[test]
repository: http://example.com/test/
username: testusername
password: testpassword
'''

        with open(self.test_pypi_rc_path, 'w+') as f:
            f.write(existing_pypi_rc)

        self.test_subject.login()

        with open(self.test_pypi_rc_path) as f:
            test_pypi_rc_str = f.read()

        self._assert_pypi_rc_has_expected_content(
            pypi_rc_str=test_pypi_rc_str,
            server='codeartifact',
            repo_url=self.endpoint,
            username='aws',
            password=self.auth_token,
        )

        self._assert_pypi_rc_has_expected_content(
            pypi_rc_str=test_pypi_rc_str,
            server='pypi',
            repo_url='http://www.python.org/pypi/',
            username='monty',
            password='JgCXIr5xGG',
        )

        self._assert_pypi_rc_has_expected_content(
            pypi_rc_str=test_pypi_rc_str,
            server='test',
            repo_url='http://example.com/test/',
            username='testusername',
            password='testpassword',
        )

    def test_login_existing_pypi_rc_with_codeartifact_not_clobbered(self):
        existing_pypi_rc = '''\
[distutils]
index-servers=
    pypi
    codeartifact

[pypi]
repository: http://www.python.org/pypi/
username: monty
password: JgCXIr5xGG

[codeartifact]
repository: https://test-testOwner.codeartifact.aws.a2z.com/pypi/testRepo/
username: aws
password: expired_token
'''

        with open(self.test_pypi_rc_path, 'w+') as f:
            f.write(existing_pypi_rc)

        self.test_subject.login()

        with open(self.test_pypi_rc_path) as f:
            test_pypi_rc_str = f.read()

        self._assert_pypi_rc_has_expected_content(
            pypi_rc_str=test_pypi_rc_str,
            server='codeartifact',
            repo_url=self.endpoint,
            username='aws',
            password=self.auth_token,
        )

        self._assert_pypi_rc_has_expected_content(
            pypi_rc_str=test_pypi_rc_str,
            server='pypi',
            repo_url='http://www.python.org/pypi/',
            username='monty',
            password='JgCXIr5xGG',
        )

    @skip_if_windows("Unix file permissions are not supported on Windows.")
    def test_login_sets_secure_permissions_on_new_file(self):
        self.assertFalse(os.path.exists(self.test_pypi_rc_path))
        self.test_subject.login()
        file_mode = os.stat(self.test_pypi_rc_path).st_mode
        self.assertEqual(stat.S_IMODE(file_mode), 0o600)

    @skip_if_windows("Unix file permissions are not supported on Windows.")
    def test_login_adjusts_permissions_on_preexisting_file(self):
        existing_pypi_rc = '''\
[distutils]
index-servers=
    pypi

[pypi]
repository: http://www.python.org/pypi/
username: test
password: JgCXIr5xGG
'''
        with open(self.test_pypi_rc_path, 'w') as f:
            f.write(existing_pypi_rc)
        os.chmod(self.test_pypi_rc_path, 0o644)
        self.test_subject.login()
        file_mode = os.stat(self.test_pypi_rc_path).st_mode
        self.assertEqual(stat.S_IMODE(file_mode), 0o600)

    @skip_if_windows("Unix file permissions are not supported on Windows.")
    @mock.patch('sys.stderr')
    @mock.patch('os.chmod', side_effect=OSError(
        errno.EPERM, 'Operation not permitted'
    ))
    def test_login_writes_error_when_chmod_fails(self, mock_chmod, mock_stderr):
        self.test_subject.login()
        mock_stderr.write.assert_any_call(
            'Unable to set file permissions for %s: '
            '[Errno 1] Operation not permitted%s'
            % (self.test_pypi_rc_path, os.linesep)
        )

    def test_login_existing_invalid_pypi_rc_error(self):
        # This is an invalid pypirc as the list of servers are expected under
        # an 'index-servers' option instead of 'servers'.
        existing_pypi_rc = '''\
[distutils]
servers=
    pypi

[pypi]
repository: http://www.python.org/pypi/
username: monty
password: JgCXIr5xGG
'''

        with open(self.test_pypi_rc_path, 'w+') as f:
            f.write(existing_pypi_rc)

        with open(self.test_pypi_rc_path) as f:
            original_content = f.read()

        with self.assertRaises(Exception):
            self.test_subject.login()

        # We should just leave the pypirc untouched when it's invalid.
        with open(self.test_pypi_rc_path) as f:
            self.assertEqual(f.read(), original_content)


class TestRelativeExpirationTime(unittest.TestCase):
    def test_with_years_months_days(self):
        remaining = relativedelta(years=1, months=9)
        message = get_relative_expiration_time(remaining)
        self.assertEqual(message, '1 year and 9 months')

    def test_with_years_months(self):
        remaining = relativedelta(
            years=1, months=8, days=30, hours=23, minutes=59, seconds=30
        )
        message = get_relative_expiration_time(remaining)
        self.assertEqual(message, '1 year and 8 months')

    def test_with_years_month(self):
        remaining = relativedelta(
            years=3, days=30, hours=23, minutes=59, seconds=30
        )
        message = get_relative_expiration_time(remaining)
        self.assertEqual(message, '3 years')

    def test_with_years_days(self):
        remaining = relativedelta(years=1, days=9)
        message = get_relative_expiration_time(remaining)
        self.assertEqual(message, '1 year')

    def test_with_year(self):
        remaining = relativedelta(months=11, days=30)
        message = get_relative_expiration_time(remaining)
        self.assertEqual(message, '11 months and 30 days')

    def test_with_years(self):
        remaining = relativedelta(years=1, months=11)
        message = get_relative_expiration_time(remaining)
        self.assertEqual(message, '1 year and 11 months')

    def test_with_years_days_hours_minutes(self):
        remaining = relativedelta(years=2, days=7, hours=11, minutes=44)
        message = get_relative_expiration_time(remaining)
        self.assertEqual(message, '2 years')

    def test_with_days_minutes(self):
        remaining = relativedelta(days=1, minutes=44)
        message = get_relative_expiration_time(remaining)
        self.assertEqual(message, '1 day')

    def test_with_day(self):
        remaining = relativedelta(days=1)
        message = get_relative_expiration_time(remaining)
        self.assertEqual(message, '1 day')

    def test_with_hour(self):
        self.expiration = (
            datetime.now(tzlocal()) + relativedelta(hours=1)
        ).replace(microsecond=0)
        remaining = relativedelta(
            self.expiration, datetime.now(tzutc())
        ) + relativedelta(seconds=30)
        message = get_relative_expiration_time(remaining)
        self.assertEqual(message, '1 hour')

    def test_with_minutes_seconds(self):
        remaining = relativedelta(hours=1)
        message = get_relative_expiration_time(remaining)
        self.assertEqual(message, '1 hour')

    def test_with_full_time(self):
        remaining = relativedelta(
            years=2, months=3, days=7, hours=11, minutes=44
        )
        message = get_relative_expiration_time(remaining)
        self.assertEqual(message, '2 years and 3 months')



import botocore.crt.auth

from tests.unit.botocore.auth.test_signers import (
    TestS3SigV4Auth,
    TestSigV4Presign,
    TestSigV4Resign,
)


class TestCrtS3SigV4Auth(TestS3SigV4Auth):
    # Repeat TestS3SigV4Auth tests, but using CRT signer
    AuthClass = botocore.crt.auth.CrtS3SigV4Auth


class TestCrtSigV4Resign(TestSigV4Resign):
    AuthClass = botocore.crt.auth.CrtSigV4Auth


class TestCrtSigV4Presign(TestSigV4Presign):
    AuthClass = botocore.crt.auth.CrtSigV4QueryAuth



# Copyright 2024 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import fnmatch
import functools
import importlib.metadata
import json
import os
import re
import site
from collections.abc import Iterator
from pathlib import Path
from typing import Dict, List, Tuple

import pytest
from packaging.requirements import Requirement

_NESTED_STR_DICT = Dict[str, "_NESTED_STR_DICT"]
_ROOT_PATH = Path(__file__).parents[2]
_LOCKFILE_PATTERN = re.compile(r'^(?P<package>.+)==(?P<version>[^\s\\]+)')


@pytest.fixture()
def awscli_package():
    return Package(name="awscli")


def parse_lockfile(lockfile: Path):
    requirements = {}
    with open(lockfile) as f:
        for line in f.readlines():
            match = _LOCKFILE_PATTERN.match(line)
            if not match:
                continue
            package = match.group("package")
            version = match.group("version")
            requirements[package] = version
    return requirements


class Package:
    def __init__(self, name: str) -> None:
        self.name = name

    @functools.cached_property
    def runtime_dependencies(self) -> "DependencyClosure":
        return self._get_runtime_closure()

    def _get_runtime_closure(self) -> "DependencyClosure":
        closure = DependencyClosure()
        for requirement in self._get_runtime_requirements():
            if self._requirement_applies_to_environment(requirement):
                closure[requirement] = Package(name=requirement.name)
        return closure

    def _get_runtime_requirements(self) -> List[Requirement]:
        req_strings = self._get_distribution(self.name).requires
        if req_strings is None:
            return []
        return [Requirement(req_string) for req_string in req_strings]

    def _requirement_applies_to_environment(
        self, requirement: Requirement
    ) -> bool:
        # Do not include any requirements defined as extras as currently
        # our dependency closure does not use any extras
        if requirement.extras:
            return False
        # Only include requirements where the markers apply to the current
        # environment.
        if requirement.marker and not requirement.marker.evaluate():
            return False
        return True

    def _get_distribution(self, name: str) -> importlib.metadata.Distribution:
        # For v2, we inject our own MetaPathFinder to handle
        # botocore/s3transfer import aliases. However for the typical
        # importlib.metadata.distribution(), it extends the built-in
        # MetaPathFinder to include its own find_distributions() method
        # to search for distribution directories. Read more here:
        # https://docs.python.org/3/library/importlib.metadata.html#extending-the-search-algorithm
        #
        # Our MetaPathFinder class does not implement this method, which
        # causes importlib.metadata.distribution() to not find the "awscli"
        # package. So instead, this helper method is implemented to locate the
        # dist-info directories based off our current site-packages
        # and explicitly provide the directory to avoid needing to use
        # MetaPathFinders and thus avoid this issue.

        # Package names may have a "-". These get converted to "_" for
        # their respective directory names in the site packages directory.
        # Directory names may have a "_" in place of a "." if they were
        # built from source. If the initial normalization fails to match,
        # we try replacing "." with "_" in the package name.
        for c in ['-', '.']:
            snake_case_name = name.replace(c, "_")
            for sitepackages in site.getsitepackages():
                for filename in os.listdir(sitepackages):
                    if fnmatch.fnmatch(
                        filename, f"{snake_case_name}-*.dist-info"
                    ):
                        return importlib.metadata.Distribution.at(
                            os.path.join(sitepackages, filename)
                        )
        raise ValueError(
            f'Could not find .dist-info directory for {snake_case_name}'
        )


class DependencyClosure:
    def __init__(self) -> None:
        self._req_to_package: Dict[Requirement, Package] = {}

    def __setitem__(self, key: Requirement, value: Package) -> None:
        self._req_to_package[key] = value

    def __getitem__(self, key: Requirement) -> Package:
        return self._req_to_package[key]

    def __delitem__(self, key: Requirement) -> None:
        del self._req_to_package[key]

    def __iter__(self) -> Iterator[Requirement]:
        return iter(self._req_to_package)

    def __len__(self) -> int:
        return len(self._req_to_package)

    def walk(self) -> Iterator[Tuple[Requirement, Package]]:
        for req, package in self._req_to_package.items():
            yield req, package
            yield from package.runtime_dependencies.walk()

    def to_dict(self) -> _NESTED_STR_DICT:
        reqs = {}
        for req, package in self._req_to_package.items():
            reqs[str(req)] = package.runtime_dependencies.to_dict()
        return reqs


class TestDependencyClosure:
    def _is_bounded_version_requirement(
        self, requirement: Requirement
    ) -> bool:
        for specifier in requirement.specifier:
            if specifier.operator in ["==", "<=", "<"]:
                return True
        return False

    def _pformat_closure(self, closure: DependencyClosure) -> str:
        return json.dumps(closure.to_dict(), sort_keys=True, indent=2)

    def test_expected_runtime_dependencies(self, awscli_package):
        expected_dependencies = {
            "awscrt",
            "colorama",
            "distro",
            "docutils",
            "jmespath",
            "prompt-toolkit",
            "python-dateutil",
            "ruamel_yaml",
            "ruamel_yaml_clib",
            "six",
            "urllib3",
            "wcwidth",
        }
        actual_dependencies = set()
        for _, package in awscli_package.runtime_dependencies.walk():
            actual_dependencies.add(package.name)
        assert actual_dependencies == expected_dependencies, (
            f"Unexpected dependency found in runtime closure: "
            f"{self._pformat_closure(awscli_package.runtime_dependencies)}"
        )

    def test_expected_unbounded_runtime_dependencies(self, awscli_package):
        expected_unbounded_dependencies = {
            "six",  # Transitive dependency from python-dateutil
        }
        all_dependencies = set()
        bounded_dependencies = set()
        for req, package in awscli_package.runtime_dependencies.walk():
            all_dependencies.add(package.name)
            if self._is_bounded_version_requirement(req):
                bounded_dependencies.add(package.name)
        actual_unbounded_dependencies = all_dependencies - bounded_dependencies
        assert (
            actual_unbounded_dependencies == expected_unbounded_dependencies
        ), (
            f"Unexpected unbounded dependency found in runtime closure: "
            f"{self._pformat_closure(awscli_package.runtime_dependencies)}"
        )

    # Test lockfiles generated with pyproject.toml as a source.
    @pytest.mark.parametrize("filename", ["requirements-docs-lock.txt"])
    def test_lockfile_pins_within_runtime_dependencies_bounds(
        self, filename, awscli_package
    ):
        lockfile = parse_lockfile(Path(_ROOT_PATH, filename))
        for req, _ in awscli_package.runtime_dependencies.walk():
            if req.name not in lockfile:
                continue
            assert req.specifier.contains(lockfile[req.name])



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License'). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the 'license' file accompanying this file. This file is
# distributed on an 'AS IS' BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import base64
import binascii
import gzip
import hashlib
import json
from argparse import Namespace
from datetime import datetime, timedelta

from awscrt.crypto import RSASignatureAlgorithm
from botocore.exceptions import ClientError
from dateutil import parser, tz

from awscli.compat import BytesIO
from awscli.customizations.cloudtrail.validation import (
    DATE_FORMAT,
    CloudTrailValidateLogs,
    DigestError,
    DigestProvider,
    DigestSignatureError,
    DigestTraverser,
    InvalidDigestFormat,
    PublicKeyProvider,
    S3ClientProvider,
    Sha256RSADigestValidator,
    assert_cloudtrail_arn_is_valid,
    create_digest_traverser,
    extract_digest_key_date,
    format_date,
    is_backfill_digest_key,
    normalize_date,
    parse_date,
)
from awscli.schema import ParameterRequiredError
from awscli.testutils import BaseAWSCommandParamsTest, mock, unittest
from tests import PublicPrivateKeyLoader

from . import get_private_key_path, get_public_key_path

START_DATE = parser.parse('20140810T000000Z')
END_DATE = parser.parse('20150810T000000Z')
TEST_ACCOUNT_ID = '123456789012'
TEST_TRAIL_ARN = f'arn:aws:cloudtrail:us-east-1:{TEST_ACCOUNT_ID}:trail/foo'
VALID_TEST_KEY = (
    'MIIBCgKCAQEAn11L2YZ9h7onug2ILi1MWyHiMRsTQjfWE+pHVRLk1QjfW'
    'hirG+lpOa8NrwQ/r7Ah5bNL6HepznOU9XTDSfmmnP97mqyc7z/upfZdS/'
    'AHhYcGaz7n6Wc/RRBU6VmiPCrAUojuSk6/GjvA8iOPFsYDuBtviXarvuL'
    'PlrT9kAd4Lb+rFfR5peEgBEkhlzc5HuWO7S0y+KunqxX6jQBnXGMtxmPB'
    'PP0FylgWGNdFtks/4YSKcgqwH0YDcawP9GGGDAeCIqPWIXDLG1jOjRRzW'
    'fCmD0iJUkz8vTsn4hq/5ZxRFE7UBAUiVcGbdnDdvVfhF9C3dQiDq3k7ad'
    'QIziLT0cShgQIDAQAB'
)
TEST_ORGANIZATION_ACCOUNT_ID = '987654321098'
TEST_ORGANIZATION_ID = 'o-12345'


def create_mock_key_provider(key_list):
    """Creates a mock key provider that yields keys for each in key_list"""
    public_keys = {}
    for k in key_list:
        public_keys[k] = {'Fingerprint': k, 'Value': 'ffaa00'}
    key_provider = mock.Mock()
    key_provider.get_public_keys.return_value = public_keys
    return key_provider


def create_scenario(actions, logs=None):
    """Creates a scenario for a stack of actions

    Each action can be "gap" meaning there is no previous link, "invalid"
    meaning we should simulate an invalid digest, "missing" meaning we
    should simulate a digest is missing from S3, "bucket_change" meaning
    it is a link but the bucket is different than the previous bucket.
    Values are popped one by one off of the list until a terminal "gap"
    action is found.
    """
    keys = [str(i) for i in range(len(actions))]
    key_provider = create_mock_key_provider(keys)
    digest_provider = MockDigestProvider(actions, logs)
    digest_validator = mock.Mock()

    def validate(bucket, key, public_key, digest_data, digest_str):
        if '_invalid' in digest_data:
            raise DigestError('invalid error')

    digest_validator.validate = validate
    return key_provider, digest_provider, digest_validator


def collecting_callback():
    """Create and return a callback and a list populated with call args"""
    calls = []

    def cb(**kwargs):
        calls.append(kwargs)

    return cb, calls


class MockDigestProvider:
    def __init__(self, actions, logs=None):
        self.logs = logs or []
        self.actions = actions
        self.calls = {'fetch_digest': [], 'load_digest_keys_in_range': []}
        self.digests = []
        self.backfill_digests = []
        self.trail_home_region = 'us-east-1'
        for i in range(len(self.actions)):
            self.digests.append(self.get_key_at_position(i))
            self.backfill_digests.append(
                self.get_key_at_position(i, is_backfill=True)
            )

    def get_key_at_position(self, position, is_backfill=False):
        """Get digest key at position, generating backfill keys if needed."""
        dt = START_DATE + timedelta(hours=position)

        if is_backfill:
            key = (
                f'AWSLogs/{TEST_ACCOUNT_ID}/CloudTrail-Digest/us-east-1/{dt.strftime("%Y/%m/%d")}/'
                f'{TEST_ACCOUNT_ID}_CloudTrail-Digest_us-east-1_foo_us-east-1_{dt.strftime(DATE_FORMAT)}_backfill.json.gz'
            )
        else:
            key = (
                f'AWSLogs/{TEST_ACCOUNT_ID}/CloudTrail-Digest/us-east-1/{dt.strftime("%Y/%m/%d")}/'
                f'{TEST_ACCOUNT_ID}_CloudTrail-Digest_us-east-1_foo_us-east-1_{dt.strftime(DATE_FORMAT)}.json.gz'
            )
        return key

    @staticmethod
    def create_digest(
        fingerprint,
        start_date,
        key,
        bucket,
        next_key=None,
        next_bucket=None,
        logs=None,
    ):
        digest_end_date = start_date + timedelta(hours=1, minutes=30)
        return {
            'digestPublicKeyFingerprint': fingerprint,
            'digestEndTime': digest_end_date.strftime(DATE_FORMAT),
            'digestStartTime': start_date.strftime(DATE_FORMAT),
            'previousDigestS3Bucket': next_bucket,
            'previousDigestS3Object': next_key,
            'digestS3Bucket': bucket,
            'digestS3Object': key,
            'awsAccountId': TEST_ACCOUNT_ID,
            'previousDigestSignature': 'abcd',
            'logFiles': logs or [],
        }

    @staticmethod
    def create_link(
        key, next_key, next_bucket, position, action, logs, bucket
    ):
        """Creates a link in a digest chain for testing."""
        digest_logs = []
        if len(logs) > position:
            digest_logs = logs[position]
        end_date = parse_date(extract_digest_key_date(key))
        # gap actions have no previous link.
        if action == 'gap':
            digest = MockDigestProvider.create_digest(
                key=key,
                bucket=bucket,
                fingerprint=str(position),
                start_date=end_date,
                logs=digest_logs,
            )
        else:
            digest = MockDigestProvider.create_digest(
                key=key,
                bucket=bucket,
                fingerprint=str(position),
                start_date=end_date,
                next_bucket=next_bucket,
                next_key=next_key,
                logs=digest_logs,
            )
            if action == 'invalid':
                digest['_invalid'] = True

        digest['_signature'] = 'mock_signature'
        digest['_signature_algorithm'] = 'SHA256'

        if is_backfill_digest_key(key):
            digest['_backfill_generation_timestamp'] = '2025-09-01T00:00:00Z'

        return digest, json.dumps(digest)

    def load_digest_keys_in_range(
        self, bucket, prefix, start_date, end_date, is_backfill=False
    ):
        self.calls['load_digest_keys_in_range'].append(locals())
        if is_backfill:
            return list(self.backfill_digests)
        return list(self.digests)

    def fetch_digest(self, bucket, key):
        self.calls['fetch_digest'].append(key)
        position = (
            self.backfill_digests.index(key)
            if is_backfill_digest_key(key)
            else self.digests.index(key)
        )
        action = self.actions[position]
        # Simulate a digest missing from S3
        if action == 'missing':
            raise ClientError(
                {'Error': {'Code': 'NoSuchKey', 'Message': 'foo'}}, 'GetObject'
            )

        next_key = self.get_key_at_position(
            position - 1, is_backfill_digest_key(key)
        )
        next_bucket = int(bucket)
        if action == 'bucket_change':
            next_bucket += 1

        return self.create_link(
            key,
            next_key,
            str(next_bucket),
            position,
            action,
            self.logs,
            bucket,
        )


class TestValidation(unittest.TestCase):
    def test_formats_dates(self):
        date = datetime(2015, 8, 21, tzinfo=tz.tzutc())
        self.assertEqual('20150821T000000Z', format_date(date))

    def test_parses_dates_with_better_error_message(self):
        try:
            parse_date('foo')
            self.fail('Should have failed to parse')
        except ValueError as e:
            self.assertIn('Unable to parse date value: foo', str(e))

    def test_parses_dates(self):
        date = parse_date('August 25, 2015 00:00:00 UTC')
        self.assertEqual(date, datetime(2015, 8, 25, tzinfo=tz.tzutc()))

    def test_ensures_cloudtrail_arns_are_valid(self):
        try:
            assert_cloudtrail_arn_is_valid('foo:bar:baz')
            self.fail('Should have failed')
        except ValueError as e:
            self.assertIn('Invalid trail ARN provided: foo:bar:baz', str(e))

    def test_ensures_cloudtrail_arns_are_valid_when_missing_resource(self):
        try:
            assert_cloudtrail_arn_is_valid(
                f'arn:aws:cloudtrail:us-east-1:{TEST_ACCOUNT_ID}:foo'
            )
            self.fail('Should have failed')
        except ValueError as e:
            self.assertIn('Invalid trail ARN provided', str(e))

    def test_allows_valid_arns(self):
        assert_cloudtrail_arn_is_valid(
            f'arn:aws:cloudtrail:us-east-1:{TEST_ACCOUNT_ID}:trail/foo'
        )

    def test_normalizes_date_timezones(self):
        date = datetime(2015, 8, 21, tzinfo=tz.tzlocal())
        normalized = normalize_date(date)
        self.assertEqual(tz.tzutc(), normalized.tzinfo)

    def test_extracts_dates_from_digest_keys(self):
        arn = (
            f'AWSLogs/{TEST_ACCOUNT_ID}/CloudTrail-Digest/us-east-1/2015/08/'
            f'16/{TEST_ACCOUNT_ID}_CloudTrail-Digest_us-east-1_foo_us-east-1_'
            '20150816T230550Z.json.gz'
        )
        self.assertEqual('20150816T230550Z', extract_digest_key_date(arn))

    def test_is_backfill_digest_key_identifies_standard_digest(self):
        standard_key = (
            f'AWSLogs/{TEST_ACCOUNT_ID}/CloudTrail-Digest/us-east-1/2015/08/'
            f'16/{TEST_ACCOUNT_ID}_CloudTrail-Digest_us-east-1_foo_us-east-1_'
            '20150816T230550Z.json.gz'
        )
        self.assertFalse(is_backfill_digest_key(standard_key))

    def test_is_backfill_digest_key_identifies_backfill_digest(self):
        backfill_key = (
            f'AWSLogs/{TEST_ACCOUNT_ID}/CloudTrail-Digest/us-east-1/2015/08/'
            f'16/{TEST_ACCOUNT_ID}_CloudTrail-Digest_us-east-1_foo_us-east-1_'
            '20150816T230550Z_backfill.json.gz'
        )
        self.assertTrue(is_backfill_digest_key(backfill_key))

    def test_is_backfill_digest_key_handles_edge_cases(self):
        # Test with different file extensions
        self.assertFalse(is_backfill_digest_key('file.txt'))
        self.assertFalse(is_backfill_digest_key('file_backfill.txt'))
        self.assertFalse(is_backfill_digest_key('file.json.gz'))

        # Test with backfill in middle of filename
        self.assertFalse(is_backfill_digest_key('file_backfill_other.json.gz'))

    def test_extracts_dates_from_backfill_digest_keys(self):
        backfill_key = (
            f'AWSLogs/{TEST_ACCOUNT_ID}/CloudTrail-Digest/us-east-1/2015/08/'
            f'16/{TEST_ACCOUNT_ID}_CloudTrail-Digest_us-east-1_foo_us-east-1_'
            '20150816T230550Z_backfill.json.gz'
        )
        self.assertEqual(
            '20150816T230550Z', extract_digest_key_date(backfill_key)
        )

    def test_extracts_dates_from_standard_digest_keys(self):
        standard_key = (
            f'AWSLogs/{TEST_ACCOUNT_ID}/CloudTrail-Digest/us-east-1/2015/08/'
            f'16/{TEST_ACCOUNT_ID}_CloudTrail-Digest_us-east-1_foo_us-east-1_'
            '20150816T230550Z.json.gz'
        )
        self.assertEqual(
            '20150816T230550Z', extract_digest_key_date(standard_key)
        )

    def test_creates_traverser(self):
        mock_s3_provider = mock.Mock()
        traverser = create_digest_traverser(
            trail_arn=TEST_TRAIL_ARN,
            cloudtrail_client=mock.Mock(),
            organization_client=mock.Mock(),
            trail_source_region='us-east-1',
            s3_client_provider=mock_s3_provider,
            bucket='bucket',
            prefix='prefix',
        )
        self.assertEqual('bucket', traverser.starting_bucket)
        self.assertEqual('prefix', traverser.starting_prefix)
        digest_provider = traverser.digest_provider
        self.assertEqual('us-east-1', digest_provider.trail_home_region)
        self.assertEqual('foo', digest_provider.trail_name)

    def test_creates_traverser_account_id(self):
        mock_s3_provider = mock.Mock()
        traverser = create_digest_traverser(
            trail_arn=TEST_TRAIL_ARN,
            cloudtrail_client=mock.Mock(),
            organization_client=mock.Mock(),
            trail_source_region='us-east-1',
            s3_client_provider=mock_s3_provider,
            bucket='bucket',
            prefix='prefix',
            account_id=TEST_ORGANIZATION_ACCOUNT_ID,
        )
        self.assertEqual('bucket', traverser.starting_bucket)
        self.assertEqual('prefix', traverser.starting_prefix)
        digest_provider = traverser.digest_provider
        self.assertEqual('us-east-1', digest_provider.trail_home_region)
        self.assertEqual('foo', digest_provider.trail_name)
        self.assertEqual(
            TEST_ORGANIZATION_ACCOUNT_ID, digest_provider.account_id
        )

    def test_creates_traverser_and_gets_trail_by_arn(self):
        cloudtrail_client = mock.Mock()
        cloudtrail_client.describe_trails.return_value = {
            'trailList': [
                {
                    'TrailARN': TEST_TRAIL_ARN,
                    'S3BucketName': 'bucket',
                    'S3KeyPrefix': 'prefix',
                    'IsOrganizationTrail': False,
                }
            ]
        }
        traverser = create_digest_traverser(
            trail_arn=TEST_TRAIL_ARN,
            trail_source_region='us-east-1',
            cloudtrail_client=cloudtrail_client,
            organization_client=mock.Mock(),
            s3_client_provider=mock.Mock(),
        )
        self.assertEqual('bucket', traverser.starting_bucket)
        self.assertEqual('prefix', traverser.starting_prefix)
        digest_provider = traverser.digest_provider
        self.assertEqual('us-east-1', digest_provider.trail_home_region)
        self.assertEqual('foo', digest_provider.trail_name)
        self.assertEqual(TEST_ACCOUNT_ID, digest_provider.account_id)

    def test_create_traverser_organizational_trail_not_launched(self):
        cloudtrail_client = mock.Mock()
        cloudtrail_client.describe_trails.return_value = {
            'trailList': [
                {
                    'TrailARN': TEST_TRAIL_ARN,
                    'S3BucketName': 'bucket',
                    'S3KeyPrefix': 'prefix',
                }
            ]
        }
        traverser = create_digest_traverser(
            trail_arn=TEST_TRAIL_ARN,
            trail_source_region='us-east-1',
            cloudtrail_client=cloudtrail_client,
            organization_client=mock.Mock(),
            s3_client_provider=mock.Mock(),
        )
        self.assertEqual('bucket', traverser.starting_bucket)
        self.assertEqual('prefix', traverser.starting_prefix)
        digest_provider = traverser.digest_provider
        self.assertEqual('us-east-1', digest_provider.trail_home_region)
        self.assertEqual('foo', digest_provider.trail_name)
        self.assertEqual(TEST_ACCOUNT_ID, digest_provider.account_id)

    def test_creates_traverser_and_gets_trail_by_arn_s3_bucket_specified(self):
        cloudtrail_client = mock.Mock()
        traverser = create_digest_traverser(
            trail_arn=TEST_TRAIL_ARN,
            trail_source_region='us-east-1',
            cloudtrail_client=cloudtrail_client,
            organization_client=mock.Mock(),
            s3_client_provider=mock.Mock(),
            bucket="bucket",
        )
        self.assertEqual('bucket', traverser.starting_bucket)
        digest_provider = traverser.digest_provider
        self.assertEqual('us-east-1', digest_provider.trail_home_region)
        self.assertEqual('foo', digest_provider.trail_name)
        self.assertEqual(TEST_ACCOUNT_ID, digest_provider.account_id)

    def test_creates_traverser_and_gets_organization_id(self):
        cloudtrail_client = mock.Mock()
        cloudtrail_client.describe_trails.return_value = {
            'trailList': [
                {
                    'TrailARN': TEST_TRAIL_ARN,
                    'S3BucketName': 'bucket',
                    'S3KeyPrefix': 'prefix',
                    'IsOrganizationTrail': True,
                }
            ]
        }
        organization_client = mock.Mock()
        organization_client.describe_organization.return_value = {
            "Organization": {
                "MasterAccountId": TEST_ACCOUNT_ID,
                "Id": TEST_ORGANIZATION_ID,
            }
        }
        traverser = create_digest_traverser(
            trail_arn=TEST_TRAIL_ARN,
            trail_source_region='us-east-1',
            cloudtrail_client=cloudtrail_client,
            organization_client=organization_client,
            s3_client_provider=mock.Mock(),
            account_id=TEST_ACCOUNT_ID,
        )
        self.assertEqual('bucket', traverser.starting_bucket)
        self.assertEqual('prefix', traverser.starting_prefix)
        digest_provider = traverser.digest_provider
        self.assertEqual('us-east-1', digest_provider.trail_home_region)
        self.assertEqual('foo', digest_provider.trail_name)
        self.assertEqual(TEST_ORGANIZATION_ID, digest_provider.organization_id)

    def test_creates_traverser_organization_trail_missing_account_id(self):
        cloudtrail_client = mock.Mock()
        cloudtrail_client.describe_trails.return_value = {
            'trailList': [
                {
                    'TrailARN': TEST_TRAIL_ARN,
                    'S3BucketName': 'bucket',
                    'S3KeyPrefix': 'prefix',
                    'IsOrganizationTrail': True,
                }
            ]
        }
        organization_client = mock.Mock()
        organization_client.describe_organization.return_value = {
            "Organization": {
                "MasterAccountId": TEST_ACCOUNT_ID,
                "Id": TEST_ORGANIZATION_ID,
            }
        }
        with self.assertRaises(ParameterRequiredError):
            create_digest_traverser(
                trail_arn=TEST_TRAIL_ARN,
                trail_source_region='us-east-1',
                cloudtrail_client=cloudtrail_client,
                organization_client=organization_client,
                s3_client_provider=mock.Mock(),
            )


class TestPublicKeyProvider(unittest.TestCase):
    def test_returns_public_key_in_range(self):
        cloudtrail_client = mock.Mock()
        cloudtrail_client.list_public_keys.return_value = {
            'PublicKeyList': [
                {'Fingerprint': 'a', 'OtherData': 'a', 'Value': 'a'},
                {'Fingerprint': 'b', 'OtherData': 'b', 'Value': 'b'},
                {'Fingerprint': 'c', 'OtherData': 'c', 'Value': 'c'},
            ]
        }
        provider = PublicKeyProvider(cloudtrail_client)
        start_date = START_DATE
        end_date = start_date + timedelta(days=2)
        keys = provider.get_public_keys(start_date, end_date)
        self.assertEqual(
            {
                'a': {'Fingerprint': 'a', 'OtherData': 'a', 'Value': 'a'},
                'b': {'Fingerprint': 'b', 'OtherData': 'b', 'Value': 'b'},
                'c': {'Fingerprint': 'c', 'OtherData': 'c', 'Value': 'c'},
            },
            keys,
        )
        cloudtrail_client.list_public_keys.assert_has_calls(
            [mock.call(EndTime=end_date, StartTime=start_date)]
        )


class TestSha256RSADigestValidator(unittest.TestCase):
    def setUp(self):
        self._digest_data = {
            'digestStartTime': 'baz',
            'digestEndTime': 'foo',
            'awsAccountId': 'account',
            'digestPublicKeyFingerprint': 'abc',
            'digestS3Bucket': 'bucket',
            'digestS3Object': 'object',
            'previousDigestSignature': 'xyz',
        }
        self._inflated_digest = json.dumps(self._digest_data).encode()
        self._digest_data['_signature'] = 'aeff'

    def test_validates_digests(self):
        (
            public_key,
            private_key,
        ) = PublicPrivateKeyLoader.load_private_key_and_public_key(
            get_private_key_path(), get_public_key_path()
        )
        sha256_hash = hashlib.sha256(self._inflated_digest)
        string_to_sign = f"{self._digest_data['digestEndTime']}\n{self._digest_data['digestS3Bucket']}/{self._digest_data['digestS3Object']}\n{sha256_hash.hexdigest()}\n{self._digest_data['previousDigestSignature']}"
        to_sign = string_to_sign.encode()
        signature = private_key.sign(
            signature_algorithm=RSASignatureAlgorithm.PKCS1_5_SHA256,
            digest=hashlib.sha256(to_sign).digest(),
        )

        self._digest_data['_signature'] = binascii.hexlify(signature)
        validator = Sha256RSADigestValidator()
        validator.validate(
            'b', 'k', public_key, self._digest_data, self._inflated_digest
        )

    def test_does_not_expose_underlying_key_decoding_error(self):
        validator = Sha256RSADigestValidator()
        try:
            validator.validate('b', 'k', 'YQo=', self._digest_data, b'invalid')
            self.fail('Should have failed')
        except DigestError as e:
            self.assertEqual(
                (
                    'Digest file\ts3://b/k\tINVALID: Unable to load '
                    'PKCS #1 key with fingerprint abc'
                ),
                str(e),
            )

    def test_does_not_expose_underlying_validation_error(self):
        validator = Sha256RSADigestValidator()
        try:
            validator.validate(
                'b', 'k', VALID_TEST_KEY, self._digest_data, b'invalid'
            )
            self.fail('Should have failed')
        except DigestSignatureError as e:
            self.assertEqual(
                (
                    'Digest file\ts3://b/k\tINVALID: signature '
                    'verification failed'
                ),
                str(e),
            )

    def test_properly_signs_when_no_previous_signature(self):
        validator = Sha256RSADigestValidator()
        digest_data = {
            'digestEndTime': 'a',
            'digestS3Bucket': 'b',
            'digestS3Object': 'c',
            'previousDigestSignature': None,
        }
        signed = validator._create_string_to_sign(digest_data, b'abc')
        self.assertEqual(
            (
                b'a\nb/c\nba7816bf8f01cfea414140de5dae2223b00361a396177a9cb410ff6'
                b'1f20015ad\nnull'
            ),
            signed,
        )


class TestDigestProvider(BaseAWSCommandParamsTest):
    def _fake_key(self, date):
        parsed = parser.parse(date)
        return (
            f'prefix/AWSLogs/{TEST_ACCOUNT_ID}/CloudTrail-Digest/us-east-1/{parsed.year}/'
            f'{parsed.month}/{parsed.day}/{TEST_ACCOUNT_ID}_CloudTrail-Digest_us-east-1_foo_'
            f'us-east-1_{date}.json.gz'
        )

    def _get_mock_provider(self, s3_client):
        mock_s3_client_provider = mock.Mock()
        mock_s3_client_provider.get_client.return_value = s3_client
        return DigestProvider(
            mock_s3_client_provider, TEST_ACCOUNT_ID, 'foo', 'us-east-1'
        )

    def test_initializes_public_properties(self):
        client = mock.Mock()
        provider = DigestProvider(client, TEST_ACCOUNT_ID, 'foo', 'us-east-1')
        self.assertEqual(TEST_ACCOUNT_ID, provider.account_id)
        self.assertEqual('foo', provider.trail_name)
        self.assertEqual('us-east-1', provider.trail_home_region)

    def test_returns_digests_in_range(self):
        s3_client = self.driver.session.create_client('s3')
        keys = [
            self._fake_key(format_date(START_DATE - timedelta(days=1))),
            self._fake_key(format_date(START_DATE + timedelta(days=1))),
            self._fake_key(format_date(START_DATE + timedelta(days=2))),
            self._fake_key(format_date(START_DATE + timedelta(days=3))),
            self._fake_key(format_date(END_DATE + timedelta(hours=1))),
            self._fake_key(format_date(END_DATE + timedelta(days=1))),
        ]
        # Create a key that looks similar but for a different trail.
        bad_name = keys[3].replace('foo', 'baz')
        # Create a key that looks similar but is from a different trail source
        # region (e.g., CloudTrail-Digest/us-west-2).
        bad_region = keys[3].replace(
            'CloudTrail-Digest/us-east-1', 'CloudTrail-Digest/us-west-2'
        )
        bad_region = bad_region.replace(
            'CloudTrail-Digest_us-east-1', 'CloudTrail-Digest_us-west-2'
        )
        self.parsed_responses = [
            {
                "Contents": [
                    {"Key": keys[0]},  # skip (date <)
                    {"Key": keys[1]},
                    {"Key": keys[2]},
                    {"Key": 'foo/baz/bar'},  # skip (regex (bogus))
                    {"Key": bad_name},  # skip (regex (trail name))
                    {"Key": bad_region},  # skip (regex (source))
                    {"Key": keys[3]},
                    {"Key": keys[4]},  # hour is +1, but keep
                    {"Key": keys[5]},  # skip (date >)
                ]
            }
        ]
        self.patch_make_request()
        provider = self._get_mock_provider(s3_client)
        digests = provider.load_digest_keys_in_range(
            'foo', 'prefix', START_DATE, END_DATE
        )
        self.assertNotIn(bad_name, digests)
        self.assertNotIn(bad_region, digests)
        self.assertEqual(keys[1], digests[0])
        self.assertEqual(keys[2], digests[1])
        self.assertEqual(keys[3], digests[2])
        self.assertEqual(keys[4], digests[3])

    def test_calls_list_objects_correctly(self):
        s3_client = mock.Mock()
        mock_paginate = s3_client.get_paginator.return_value.paginate
        mock_search = mock_paginate.return_value.search
        mock_search.return_value = []
        provider = self._get_mock_provider(s3_client)
        provider.load_digest_keys_in_range('1', 'prefix', START_DATE, END_DATE)
        marker = (
            'prefix/AWSLogs/{account}/CloudTrail-Digest/us-east-1/'
            '2014/08/09/{account}_CloudTrail-Digest_us-east-1_foo_'
            'us-east-1_20140809T235900Z.json.gz'
        )
        prefix = 'prefix/AWSLogs/{account}/CloudTrail-Digest/us-east-1'
        mock_paginate.assert_called_once_with(
            Bucket='1',
            Marker=marker.format(account=TEST_ACCOUNT_ID),
            Prefix=prefix.format(account=TEST_ACCOUNT_ID),
        )

    def test_calls_list_objects_correctly_org_trails(self):
        s3_client = mock.Mock()
        mock_s3_client_provider = mock.Mock()
        mock_paginate = s3_client.get_paginator.return_value.paginate
        mock_search = mock_paginate.return_value.search
        mock_search.return_value = []
        mock_s3_client_provider.get_client.return_value = s3_client
        provider = DigestProvider(
            mock_s3_client_provider,
            TEST_ORGANIZATION_ACCOUNT_ID,
            'foo',
            'us-east-1',
            'us-east-1',
            TEST_ORGANIZATION_ID,
        )
        provider.load_digest_keys_in_range('1', 'prefix', START_DATE, END_DATE)
        marker = (
            'prefix/AWSLogs/{organization_id}/{member_account}/'
            'CloudTrail-Digest/us-east-1/'
            '2014/08/09/{member_account}_CloudTrail-Digest_us-east-1_foo_'
            'us-east-1_20140809T235900Z.json.gz'
        )
        prefix = (
            'prefix/AWSLogs/{organization_id}/{member_account}/'
            'CloudTrail-Digest/us-east-1'
        )
        mock_paginate.assert_called_once_with(
            Bucket='1',
            Marker=marker.format(
                member_account=TEST_ORGANIZATION_ACCOUNT_ID,
                organization_id=TEST_ORGANIZATION_ID,
            ),
            Prefix=prefix.format(
                member_account=TEST_ORGANIZATION_ACCOUNT_ID,
                organization_id=TEST_ORGANIZATION_ID,
            ),
        )

    def test_create_digest_prefix_without_key_prefix(self):
        mock_s3_client_provider = mock.Mock()
        provider = DigestProvider(
            mock_s3_client_provider, TEST_ACCOUNT_ID, 'foo', 'us-east-1'
        )
        prefix = provider._create_digest_prefix(START_DATE, None)
        expected = f'AWSLogs/{TEST_ACCOUNT_ID}/CloudTrail-Digest/us-east-1'
        self.assertEqual(expected, prefix)

    def test_create_digest_prefix_with_key_prefix(self):
        mock_s3_client_provider = mock.Mock()
        provider = DigestProvider(
            mock_s3_client_provider, TEST_ACCOUNT_ID, 'foo', 'us-east-1'
        )
        prefix = provider._create_digest_prefix(START_DATE, 'my-prefix')
        expected = (
            f'my-prefix/AWSLogs/{TEST_ACCOUNT_ID}/CloudTrail-Digest/us-east-1'
        )
        self.assertEqual(expected, prefix)

    def test_create_digest_prefix_org_trail(self):
        mock_s3_client_provider = mock.Mock()
        provider = DigestProvider(
            mock_s3_client_provider,
            TEST_ORGANIZATION_ACCOUNT_ID,
            'foo',
            'us-east-1',
            'us-east-1',
            TEST_ORGANIZATION_ID,
        )
        prefix = provider._create_digest_prefix(START_DATE, None)
        expected = f'AWSLogs/{TEST_ORGANIZATION_ID}/{TEST_ORGANIZATION_ACCOUNT_ID}/CloudTrail-Digest/us-east-1'
        self.assertEqual(expected, prefix)

    def test_create_digest_prefix_org_trail_with_key_prefix(self):
        mock_s3_client_provider = mock.Mock()
        provider = DigestProvider(
            mock_s3_client_provider,
            TEST_ORGANIZATION_ACCOUNT_ID,
            'foo',
            'us-east-1',
            'us-east-1',
            TEST_ORGANIZATION_ID,
        )
        prefix = provider._create_digest_prefix(START_DATE, 'custom-prefix')
        expected = f'custom-prefix/AWSLogs/{TEST_ORGANIZATION_ID}/{TEST_ORGANIZATION_ACCOUNT_ID}/CloudTrail-Digest/us-east-1'
        self.assertEqual(expected, prefix)

    def test_ensures_digest_has_proper_metadata(self):
        out = BytesIO()
        f = gzip.GzipFile(fileobj=out, mode="wb")
        f.write(b'{"foo":"bar"}')
        f.close()
        gzipped_data = out.getvalue()
        s3_client = mock.Mock()
        s3_client.get_object.return_value = {
            'Body': BytesIO(gzipped_data),
            'Metadata': {},
        }
        provider = self._get_mock_provider(s3_client)
        with self.assertRaises(DigestSignatureError):
            provider.fetch_digest('bucket', 'key')

    def test_ensures_digest_can_be_gzip_inflated(self):
        s3_client = mock.Mock()
        s3_client.get_object.return_value = {
            'Body': BytesIO(b'foo'),
            'Metadata': {},
        }
        provider = self._get_mock_provider(s3_client)
        with self.assertRaises(InvalidDigestFormat):
            provider.fetch_digest('bucket', 'key')

    def test_ensures_digests_can_be_json_parsed(self):
        json_str = '{{{'
        out = BytesIO()
        f = gzip.GzipFile(fileobj=out, mode="wb")
        f.write(json_str.encode())
        f.close()
        gzipped_data = out.getvalue()
        s3_client = mock.Mock()
        s3_client.get_object.return_value = {
            'Body': BytesIO(gzipped_data),
            'Metadata': {'signature': 'abc', 'signature-algorithm': 'SHA256'},
        }
        provider = self._get_mock_provider(s3_client)
        with self.assertRaises(InvalidDigestFormat):
            provider.fetch_digest('bucket', 'key')

    def test_fetches_digests(self):
        json_str = '{"foo":"bar"}'
        out = BytesIO()
        f = gzip.GzipFile(fileobj=out, mode="wb")
        f.write(json_str.encode())
        f.close()
        gzipped_data = out.getvalue()
        s3_client = mock.Mock()
        s3_client.get_object.return_value = {
            'Body': BytesIO(gzipped_data),
            'Metadata': {'signature': 'abc', 'signature-algorithm': 'SHA256'},
        }
        provider = self._get_mock_provider(s3_client)
        result = provider.fetch_digest('bucket', 'key')
        self.assertEqual(
            {
                'foo': 'bar',
                '_signature': 'abc',
                '_signature_algorithm': 'SHA256',
            },
            result[0],
        )
        self.assertEqual(json_str.encode(), result[1])

    def _fake_backfill_key(self, date):
        parsed = parser.parse(date)
        return (
            f'prefix/AWSLogs/{TEST_ACCOUNT_ID}/CloudTrail-Digest/us-east-1/{parsed.year}/'
            f'{parsed.month}/{parsed.day}/{TEST_ACCOUNT_ID}_CloudTrail-Digest_us-east-1_foo_'
            f'us-east-1_{date}_backfill.json.gz'
        )

    def test_load_all_digest_keys_in_range_separates_standard_and_backfill(
        self,
    ):
        s3_client = self.driver.session.create_client('s3')
        standard_keys = [
            self._fake_key(format_date(START_DATE + timedelta(days=1))),
            self._fake_key(format_date(START_DATE + timedelta(days=2))),
        ]
        backfill_keys = [
            self._fake_backfill_key(
                format_date(START_DATE + timedelta(days=1))
            ),
            self._fake_backfill_key(
                format_date(START_DATE + timedelta(days=2))
            ),
        ]

        self.parsed_responses = [
            {
                "Contents": [
                    {"Key": standard_keys[0]},
                    {"Key": backfill_keys[0]},
                    {"Key": standard_keys[1]},
                    {"Key": backfill_keys[1]},
                ]
            }
        ]
        self.patch_make_request()
        provider = self._get_mock_provider(s3_client)

        standard_digests, backfill_digests = (
            provider.load_all_digest_keys_in_range(
                'foo', 'prefix', START_DATE, END_DATE
            )
        )

        self.assertEqual(standard_keys, standard_digests)
        self.assertEqual(backfill_keys, backfill_digests)

    def test_load_digest_keys_in_range_uses_cache(self):
        s3_client = mock.Mock()
        mock_paginate = s3_client.get_paginator.return_value.paginate
        mock_search = mock_paginate.return_value.search
        mock_search.return_value = []
        provider = self._get_mock_provider(s3_client)

        provider.load_digest_keys_in_range(
            'bucket', 'prefix', START_DATE, END_DATE, is_backfill=False
        )
        self.assertEqual(1, mock_paginate.call_count)

        provider.load_digest_keys_in_range(
            'bucket', 'prefix', START_DATE, END_DATE, is_backfill=True
        )
        self.assertEqual(1, mock_paginate.call_count)

        provider.load_digest_keys_in_range(
            'bucket', 'prefix', START_DATE, START_DATE, is_backfill=False
        )
        self.assertEqual(2, mock_paginate.call_count)

    def test_fetches_backfill_digests_with_metadata(self):
        json_str = '{"foo":"bar"}'
        out = BytesIO()
        f = gzip.GzipFile(fileobj=out, mode="wb")
        f.write(json_str.encode())
        f.close()
        gzipped_data = out.getvalue()
        s3_client = mock.Mock()
        s3_client.get_object.return_value = {
            'Body': BytesIO(gzipped_data),
            'Metadata': {
                'signature': 'abc',
                'signature-algorithm': 'SHA256',
                'backfill-generation-timestamp': '2025-09-01T00:00:00Z',
            },
        }
        provider = self._get_mock_provider(s3_client)
        backfill_key = self._fake_backfill_key(format_date(START_DATE))

        result = provider.fetch_digest('bucket', backfill_key)

        self.assertEqual(
            {
                'foo': 'bar',
                '_signature': 'abc',
                '_signature_algorithm': 'SHA256',
                '_backfill_generation_timestamp': '2025-09-01T00:00:00Z',
            },
            result[0],
        )
        self.assertEqual(json_str.encode(), result[1])

    def test_ensures_backfill_digest_has_proper_metadata(self):
        json_str = '{"foo":"bar"}'
        out = BytesIO()
        f = gzip.GzipFile(fileobj=out, mode="wb")
        f.write(json_str.encode())
        f.close()
        gzipped_data = out.getvalue()
        s3_client = mock.Mock()
        s3_client.get_object.return_value = {
            'Body': BytesIO(gzipped_data),
            'Metadata': {
                'signature': 'abc',
                'signature-algorithm': 'SHA256',
                # Missing backfill-generation-timestamp
            },
        }
        provider = self._get_mock_provider(s3_client)
        backfill_key = self._fake_backfill_key(format_date(START_DATE))

        with self.assertRaises(InvalidDigestFormat):
            provider.fetch_digest('bucket', backfill_key)


class TestDigestTraverser(unittest.TestCase):
    def test_initializes_with_default_validator(self):
        provider = mock.Mock()
        traverser = DigestTraverser(
            digest_provider=provider,
            starting_bucket='1',
            starting_prefix='baz',
            public_key_provider=mock.Mock(),
        )
        self.assertEqual('1', traverser.starting_bucket)
        self.assertEqual('baz', traverser.starting_prefix)
        self.assertEqual(provider, traverser.digest_provider)

    def test_ensures_public_keys_are_loaded(self):
        start_date = START_DATE
        end_date = END_DATE
        digest_provider = mock.Mock()
        key_provider = mock.Mock()
        key_provider.get_public_keys.return_value = []
        traverser = DigestTraverser(
            digest_provider=digest_provider,
            starting_bucket='1',
            starting_prefix='baz',
            public_key_provider=key_provider,
        )
        digest_iter = traverser.traverse_digests(start_date, end_date)
        with self.assertRaises(RuntimeError):
            next(digest_iter)
        key_provider.get_public_keys.assert_called_with(
            start_date, end_date + timedelta(hours=2)
        )

    def test_ensures_public_key_is_found(self):
        start_date = START_DATE
        end_date = END_DATE
        key_name = end_date.strftime(DATE_FORMAT) + '.json.gz'
        region = 'us-west-2'
        digest_provider = mock.Mock()
        digest_provider.trail_home_region = region
        digest_provider.load_digest_keys_in_range.return_value = [key_name]
        digest_provider.fetch_digest.return_value = (
            {
                'digestEndTime': 'foo',
                'digestStartTime': 'foo',
                'awsAccountId': 'account',
                'digestPublicKeyFingerprint': 'abc',
                'digestS3Bucket': '1',
                'digestS3Object': key_name,
                'previousDigestSignature': 'xyz',
            },
            'abc',
        )
        key_provider = mock.Mock()
        key_provider.get_public_keys.return_value = [{'Fingerprint': 'a'}]
        on_invalid, calls = collecting_callback()
        traverser = DigestTraverser(
            digest_provider=digest_provider,
            starting_bucket='1',
            starting_prefix='baz',
            public_key_provider=key_provider,
            on_invalid=on_invalid,
        )
        digest_iter = traverser.traverse_digests(start_date, end_date)
        with self.assertRaises(StopIteration):
            next(digest_iter)
        self.assertEqual(1, len(calls))
        self.assertEqual(
            f'Digest file\ts3://1/{key_name}\tINVALID: public key not '
            f'found in region {region} for fingerprint abc',
            calls[0]['message'],
        )

    def test_invokes_digest_validator(self):
        start_date = START_DATE
        end_date = END_DATE
        key_name = end_date.strftime(DATE_FORMAT) + '.json.gz'
        digest = {
            'digestPublicKeyFingerprint': 'a',
            'digestS3Bucket': '1',
            'digestS3Object': key_name,
            'previousDigestSignature': '...',
            'digestStartTime': (end_date - timedelta(hours=1)).strftime(
                DATE_FORMAT
            ),
            'digestEndTime': end_date.strftime(DATE_FORMAT),
        }
        digest_provider = mock.Mock()
        digest_provider.load_digest_keys_in_range.return_value = [key_name]
        digest_provider.fetch_digest.return_value = (digest, key_name)
        key_provider = mock.Mock()
        public_keys = {'a': {'Fingerprint': 'a', 'Value': 'a'}}
        key_provider.get_public_keys.return_value = public_keys
        digest_validator = mock.Mock()
        traverser = DigestTraverser(
            digest_provider=digest_provider,
            starting_bucket='1',
            starting_prefix='baz',
            public_key_provider=key_provider,
            digest_validator=digest_validator,
        )
        digest_iter = traverser.traverse_digests(start_date, end_date)
        self.assertEqual(digest, next(digest_iter))
        digest_validator.validate.assert_called_with(
            '1', key_name, public_keys['a']['Value'], digest, key_name
        )

    def test_ensures_digest_from_same_location_as_json_contents(self):
        start_date = START_DATE
        end_date = END_DATE
        callback, collected = collecting_callback()
        key_name = end_date.strftime(DATE_FORMAT) + '.json.gz'
        digest = {
            'digestPublicKeyFingerprint': 'a',
            'digestS3Bucket': 'not_same',
            'digestS3Object': key_name,
            'digestEndTime': end_date.strftime(DATE_FORMAT),
        }
        digest_provider = mock.Mock()
        digest_provider.load_digest_keys_in_range.return_value = [key_name]
        digest_provider.fetch_digest.return_value = (digest, key_name)
        key_provider = mock.Mock()
        digest_validator = mock.Mock()
        traverser = DigestTraverser(
            digest_provider=digest_provider,
            starting_bucket='1',
            starting_prefix='baz',
            public_key_provider=key_provider,
            digest_validator=digest_validator,
            on_invalid=callback,
        )
        digest_iter = traverser.traverse_digests(start_date, end_date)
        self.assertIsNone(next(digest_iter, None))
        self.assertEqual(1, len(collected))
        self.assertEqual(
            f'Digest file\ts3://1/{key_name}\tINVALID: invalid format',
            collected[0]['message'],
        )

    def test_loads_digests_in_range(self):
        start_date = START_DATE
        end_date = START_DATE + timedelta(hours=5)
        key_provider, digest_provider, validator = create_scenario(
            ['gap', 'link', 'link', 'link']
        )
        traverser = DigestTraverser(
            digest_provider=digest_provider,
            starting_bucket='1',
            starting_prefix='baz',
            public_key_provider=key_provider,
            digest_validator=validator,
        )
        collected = list(traverser.traverse_digests(start_date, end_date))
        self.assertEqual(1, key_provider.get_public_keys.call_count)
        self.assertEqual(
            1, len(digest_provider.calls['load_digest_keys_in_range'])
        )
        self.assertEqual(4, len(digest_provider.calls['fetch_digest']))
        self.assertEqual(4, len(collected))

    def test_invokes_cb_and_continues_when_missing(self):
        start_date = START_DATE
        end_date = END_DATE
        key_provider, digest_provider, validator = create_scenario(
            ['gap', 'link', 'missing', 'link']
        )
        on_missing, missing_calls = collecting_callback()
        traverser = DigestTraverser(
            digest_provider=digest_provider,
            starting_bucket='1',
            starting_prefix='baz',
            public_key_provider=key_provider,
            digest_validator=validator,
            on_missing=on_missing,
        )
        collected = list(traverser.traverse_digests(start_date, end_date))
        self.assertEqual(3, len(collected))
        self.assertEqual(1, key_provider.get_public_keys.call_count)
        self.assertEqual(1, len(missing_calls))
        # Ensure the keys were provided in the correct order.
        self.assertIn('bucket', missing_calls[0])
        self.assertIn('next_end_date', missing_calls[0])
        # Ensure the keys were provided in the correct order.
        self.assertEqual(
            digest_provider.digests[1], missing_calls[0]['next_key']
        )
        self.assertEqual(
            digest_provider.digests[2], missing_calls[0]['last_key']
        )
        # Ensure the provider was called correctly
        self.assertEqual(1, key_provider.get_public_keys.call_count)
        self.assertEqual(
            1, len(digest_provider.calls['load_digest_keys_in_range'])
        )
        self.assertEqual(4, len(digest_provider.calls['fetch_digest']))

    def test_invokes_cb_and_continues_when_invalid(self):
        start_date = START_DATE
        end_date = END_DATE
        key_provider, digest_provider, validator = create_scenario(
            ['gap', 'link', 'invalid', 'link', 'invalid']
        )
        on_invalid, invalid_calls = collecting_callback()
        traverser = DigestTraverser(
            digest_provider=digest_provider,
            starting_bucket='1',
            starting_prefix='baz',
            public_key_provider=key_provider,
            digest_validator=validator,
            on_invalid=on_invalid,
        )
        collected = list(traverser.traverse_digests(start_date, end_date))
        self.assertEqual(3, len(collected))
        self.assertEqual(1, key_provider.get_public_keys.call_count)
        self.assertEqual(2, len(invalid_calls))
        # Ensure it was invoked with all the kwargs we expected.
        self.assertIn('bucket', invalid_calls[0])
        self.assertIn('next_end_date', invalid_calls[0])
        # Ensure the keys were provided in the correct order.
        self.assertEqual(
            digest_provider.digests[4], invalid_calls[0]['last_key']
        )
        self.assertEqual(
            digest_provider.digests[3], invalid_calls[0]['next_key']
        )
        self.assertEqual(
            digest_provider.digests[2], invalid_calls[1]['last_key']
        )
        self.assertEqual(
            digest_provider.digests[1], invalid_calls[1]['next_key']
        )
        # Ensure the provider was called correctly
        self.assertEqual(1, key_provider.get_public_keys.call_count)
        self.assertEqual(
            1, len(digest_provider.calls['load_digest_keys_in_range'])
        )
        self.assertEqual(5, len(digest_provider.calls['fetch_digest']))

    def test_invokes_cb_and_continues_when_gap(self):
        start_date = START_DATE
        end_date = END_DATE
        key_provider, digest_provider, validator = create_scenario(
            ['gap', 'link', 'gap', 'gap']
        )
        on_gap, gap_calls = collecting_callback()
        traverser = DigestTraverser(
            digest_provider=digest_provider,
            starting_bucket='1',
            starting_prefix='baz',
            public_key_provider=key_provider,
            digest_validator=validator,
            on_gap=on_gap,
        )
        collected = list(traverser.traverse_digests(start_date, end_date))
        self.assertEqual(4, len(collected))
        self.assertEqual(1, key_provider.get_public_keys.call_count)
        self.assertEqual(2, len(gap_calls))
        # Ensure it was invoked with all the kwargs we expected.
        self.assertIn('bucket', gap_calls[0])
        self.assertIn('next_key', gap_calls[0])
        self.assertIn('next_end_date', gap_calls[0])
        self.assertIn('last_key', gap_calls[0])
        self.assertIn('last_start_date', gap_calls[0])
        # Ensure the keys were provided in the correct order.
        self.assertEqual(digest_provider.digests[3], gap_calls[0]['last_key'])
        self.assertEqual(digest_provider.digests[2], gap_calls[0]['next_key'])
        self.assertEqual(digest_provider.digests[2], gap_calls[1]['last_key'])
        self.assertEqual(digest_provider.digests[1], gap_calls[1]['next_key'])
        # Ensure the provider was called correctly
        self.assertEqual(1, key_provider.get_public_keys.call_count)
        self.assertEqual(
            1, len(digest_provider.calls['load_digest_keys_in_range'])
        )
        self.assertEqual(4, len(digest_provider.calls['fetch_digest']))

    def test_reloads_objects_on_bucket_change(self):
        start_date = START_DATE
        end_date = END_DATE
        key_provider, digest_provider, validator = create_scenario(
            ['gap', 'link', 'bucket_change', 'link']
        )
        traverser = DigestTraverser(
            digest_provider=digest_provider,
            starting_bucket='1',
            starting_prefix='baz',
            public_key_provider=key_provider,
            digest_validator=validator,
        )
        collected = list(traverser.traverse_digests(start_date, end_date))
        self.assertEqual(4, len(collected))
        self.assertEqual(1, key_provider.get_public_keys.call_count)
        # Ensure the provider was called correctly
        self.assertEqual(1, key_provider.get_public_keys.call_count)
        self.assertEqual(
            2, len(digest_provider.calls['load_digest_keys_in_range'])
        )
        self.assertEqual(
            ['1', '1', '2', '2'], [c['digestS3Bucket'] for c in collected]
        )

    def test_does_not_hard_fail_on_invalid_signature(self):
        start_date = START_DATE
        end_date = END_DATE
        end_timestamp = end_date.strftime(DATE_FORMAT) + '.json.gz'
        digest = {
            'digestPublicKeyFingerprint': 'a',
            'digestS3Bucket': '1',
            'digestS3Object': end_timestamp,
            'previousDigestSignature': '...',
            'digestStartTime': (end_date - timedelta(hours=1)).strftime(
                DATE_FORMAT
            ),
            'digestEndTime': end_timestamp,
            '_signature': '123',
        }
        digest_provider = mock.Mock()
        digest_provider.load_digest_keys_in_range.return_value = [
            end_timestamp
        ]
        digest_provider.fetch_digest.return_value = (digest, end_timestamp)
        key_provider = mock.Mock()
        public_keys = {'a': {'Fingerprint': 'a', 'Value': 'a'}}
        key_provider.get_public_keys.return_value = public_keys
        digest_validator = Sha256RSADigestValidator()
        on_invalid, calls = collecting_callback()
        traverser = DigestTraverser(
            digest_provider=digest_provider,
            starting_bucket='1',
            starting_prefix='baz',
            public_key_provider=key_provider,
            digest_validator=digest_validator,
            on_invalid=on_invalid,
        )
        digest_iter = traverser.traverse_digests(start_date, end_date)
        next(digest_iter, None)
        self.assertIn(
            f'Digest file\ts3://1/{end_timestamp}\tINVALID: ',
            calls[0]['message'],
        )

    def test_traverse_backfill_digests_basic(self):
        start_date = START_DATE
        end_date = START_DATE + timedelta(hours=4)
        key_provider, digest_provider, validator = create_scenario(
            ['gap', 'link', 'link', 'link']
        )

        traverser = DigestTraverser(
            digest_provider=digest_provider,
            starting_bucket='1',
            starting_prefix='baz',
            public_key_provider=key_provider,
            digest_validator=validator,
        )

        collected = list(
            traverser.traverse_digests(start_date, end_date, True)
        )

        self.assertEqual(4, len(collected))
        self.assertEqual(1, key_provider.get_public_keys.call_count)
        self.assertEqual(
            1, len(digest_provider.calls['load_digest_keys_in_range'])
        )
        self.assertEqual(4, len(digest_provider.calls['fetch_digest']))

    def test_traverse_backfill_digests_with_missing(self):
        start_date = START_DATE
        end_date = START_DATE + timedelta(hours=4)
        key_provider, digest_provider, validator = create_scenario(
            ['gap', 'link', 'missing', 'link']
        )

        on_missing, missing_calls = collecting_callback()
        traverser = DigestTraverser(
            digest_provider=digest_provider,
            starting_bucket='1',
            starting_prefix='baz',
            public_key_provider=key_provider,
            digest_validator=validator,
            on_missing=on_missing,
        )

        collected = list(
            traverser.traverse_digests(start_date, end_date, True)
        )

        self.assertEqual(3, len(collected))
        self.assertEqual(1, key_provider.get_public_keys.call_count)
        self.assertEqual(1, len(missing_calls))
        self.assertIn('bucket', missing_calls[0])
        self.assertIn('next_end_date', missing_calls[0])

    def test_traverse_backfill_digests_with_invalid(self):
        start_date = START_DATE
        end_date = START_DATE + timedelta(hours=5)
        key_provider, digest_provider, validator = create_scenario(
            ['gap', 'link', 'invalid', 'link', 'invalid']
        )

        on_invalid, invalid_calls = collecting_callback()
        traverser = DigestTraverser(
            digest_provider=digest_provider,
            starting_bucket='1',
            starting_prefix='baz',
            public_key_provider=key_provider,
            digest_validator=validator,
            on_invalid=on_invalid,
        )

        collected = list(
            traverser.traverse_digests(start_date, end_date, True)
        )

        self.assertEqual(3, len(collected))
        self.assertEqual(1, key_provider.get_public_keys.call_count)
        self.assertEqual(2, len(invalid_calls))

    def test_traverse_backfill_digests_with_gaps(self):
        start_date = START_DATE
        end_date = START_DATE + timedelta(hours=4)
        key_provider, digest_provider, validator = create_scenario(
            ['gap', 'link', 'gap', 'gap']
        )

        on_gap, gap_calls = collecting_callback()
        traverser = DigestTraverser(
            digest_provider=digest_provider,
            starting_bucket='1',
            starting_prefix='baz',
            public_key_provider=key_provider,
            digest_validator=validator,
            on_gap=on_gap,
        )

        collected = list(
            traverser.traverse_digests(start_date, end_date, True)
        )

        self.assertEqual(4, len(collected))
        self.assertEqual(1, key_provider.get_public_keys.call_count)
        self.assertEqual(2, len(gap_calls))
        for gap_call in gap_calls:
            self.assertIn('bucket', gap_call)
            self.assertIn('next_key', gap_call)

    def test_traverse_backfill_digests_bucket_change(self):
        start_date = START_DATE
        end_date = START_DATE + timedelta(hours=4)
        key_provider, digest_provider, validator = create_scenario(
            ['gap', 'link', 'bucket_change', 'link']
        )

        traverser = DigestTraverser(
            digest_provider=digest_provider,
            starting_bucket='1',
            starting_prefix='baz',
            public_key_provider=key_provider,
            digest_validator=validator,
        )

        collected = list(
            traverser.traverse_digests(start_date, end_date, True)
        )

        self.assertEqual(4, len(collected))
        self.assertEqual(1, key_provider.get_public_keys.call_count)
        self.assertEqual(
            2, len(digest_provider.calls['load_digest_keys_in_range'])
        )
        self.assertEqual(
            ['1', '1', '2', '2'], [c['digestS3Bucket'] for c in collected]
        )

    def test_traverse_mixed_standard_and_backfill_digests(self):
        start_date = START_DATE
        end_date = START_DATE + timedelta(hours=3)
        key_provider, digest_provider, validator = create_scenario(
            ['gap', 'link', 'link']
        )

        traverser = DigestTraverser(
            digest_provider=digest_provider,
            starting_bucket='1',
            starting_prefix='baz',
            public_key_provider=key_provider,
            digest_validator=validator,
        )

        standard_digests = list(
            traverser.traverse_digests(start_date, end_date)
        )
        backfill_digests = list(
            traverser.traverse_digests(start_date, end_date, True)
        )

        self.assertEqual(3, len(standard_digests))
        self.assertEqual(3, len(backfill_digests))
        self.assertEqual(2, key_provider.get_public_keys.call_count)
        self.assertEqual(
            2, len(digest_provider.calls['load_digest_keys_in_range'])
        )
        self.assertEqual(6, len(digest_provider.calls['fetch_digest']))

    def test_traverse_backfill_digests_cache_miss_triggers_multiple_api_calls(
        self,
    ):
        start_date = START_DATE
        end_date = START_DATE + timedelta(hours=3)
        key_provider, digest_provider, validator = create_scenario(
            ['gap', 'link', 'link']
        )

        call_count = 0

        def mock_get_public_keys(start_date, end_date):
            nonlocal call_count
            call_count += 1
            if call_count == 1:
                return {'2': {'Fingerprint': '2', 'Value': 'ffaa02'}}
            elif call_count == 2:
                return {'1': {'Fingerprint': '1', 'Value': 'ffaa01'}}
            else:
                return {'0': {'Fingerprint': '0', 'Value': 'ffaa00'}}

        key_provider.get_public_keys.side_effect = mock_get_public_keys

        traverser = DigestTraverser(
            digest_provider=digest_provider,
            starting_bucket='1',
            starting_prefix='baz',
            public_key_provider=key_provider,
            digest_validator=validator,
        )

        collected = list(
            traverser.traverse_digests(start_date, end_date, True)
        )

        self.assertEqual(3, len(collected))
        self.assertEqual(3, key_provider.get_public_keys.call_count)
        self.assertEqual(
            1, len(digest_provider.calls['load_digest_keys_in_range'])
        )
        self.assertEqual(3, len(digest_provider.calls['fetch_digest']))


class TestCloudTrailCommand(BaseAWSCommandParamsTest):
    def test_s3_client_created_lazily(self):
        session = mock.Mock()
        command = CloudTrailValidateLogs(session)
        parsed_globals = mock.Mock(
            region=None, verify_ssl=None, endpoint_url=None
        )
        command.setup_services(parsed_globals)
        create_client_calls = session.create_client.call_args_list
        self.assertEqual(
            create_client_calls,
            [
                mock.call('organizations', verify=None, region_name=None),
                mock.call('cloudtrail', verify=None, region_name=None),
            ],
        )

    def test_endpoint_url_is_used_for_cloudtrail(self):
        endpoint_url = 'https://mycloudtrail.aws.amazon.com/'
        session = mock.Mock()
        command = CloudTrailValidateLogs(session)
        parsed_globals = mock.Mock(
            region='foo', verify_ssl=None, endpoint_url=endpoint_url
        )
        command.setup_services(parsed_globals)
        create_client_calls = session.create_client.call_args_list
        self.assertEqual(
            create_client_calls,
            [
                mock.call('organizations', verify=None, region_name='foo'),
                # Here we should inject the endpoint_url only for cloudtrail.
                mock.call(
                    'cloudtrail',
                    verify=None,
                    region_name='foo',
                    endpoint_url=endpoint_url,
                ),
            ],
        )

    def test_initializes_args(self):
        session = mock.Mock()
        command = CloudTrailValidateLogs(session)
        start_date = START_DATE.strftime(DATE_FORMAT)
        args = Namespace(
            trail_arn='abc',
            verbose=True,
            start_time=start_date,
            s3_bucket='bucket',
            s3_prefix='prefix',
            end_time=None,
            account_id=None,
        )
        command.handle_args(args)
        self.assertEqual('abc', command.trail_arn)
        self.assertEqual(True, command.is_verbose)
        self.assertEqual('bucket', command.s3_bucket)
        self.assertEqual('prefix', command.s3_prefix)
        self.assertEqual(start_date, command.start_time.strftime(DATE_FORMAT))
        self.assertIsNotNone(command.end_time)
        self.assertGreater(command.end_time, command.start_time)


class TestS3ClientProvider(BaseAWSCommandParamsTest):
    def test_creates_clients_for_buckets_in_us_east_1(self):
        session = mock.Mock()
        s3_client = mock.Mock()
        session.create_client.return_value = s3_client
        s3_client.get_bucket_location.return_value = {'LocationConstraint': ''}
        provider = S3ClientProvider(session)
        created_client = provider.get_client('foo')
        self.assertEqual(s3_client, created_client)
        create_client_calls = session.create_client.call_args_list
        self.assertEqual(
            create_client_calls, [mock.call('s3', region_name='us-east-1')]
        )
        self.assertEqual(1, s3_client.get_bucket_location.call_count)

    def test_creates_clients_for_buckets_outside_us_east_1(self):
        session = mock.Mock()
        s3_client = mock.Mock()
        session.create_client.return_value = s3_client
        s3_client.get_bucket_location.return_value = {
            'LocationConstraint': 'us-west-2'
        }
        provider = S3ClientProvider(session, 'us-west-1')
        created_client = provider.get_client('foo')
        self.assertEqual(s3_client, created_client)
        create_client_calls = session.create_client.call_args_list
        self.assertEqual(
            create_client_calls,
            [
                mock.call('s3', region_name='us-west-1'),
                mock.call('s3', region_name='us-west-2'),
            ],
        )
        self.assertEqual(1, s3_client.get_bucket_location.call_count)

    def test_caches_previously_loaded_bucket_regions(self):
        session = mock.Mock()
        s3_client = mock.Mock()
        session.create_client.return_value = s3_client
        s3_client.get_bucket_location.return_value = {'LocationConstraint': ''}
        provider = S3ClientProvider(session)
        provider.get_client('foo')
        self.assertEqual(1, s3_client.get_bucket_location.call_count)
        provider.get_client('foo')
        self.assertEqual(1, s3_client.get_bucket_location.call_count)
        provider.get_client('bar')
        self.assertEqual(2, s3_client.get_bucket_location.call_count)
        provider.get_client('bar')
        self.assertEqual(2, s3_client.get_bucket_location.call_count)

    def test_caches_previously_loaded_clients(self):
        session = mock.Mock()
        s3_client = mock.Mock()
        session.create_client.return_value = s3_client
        s3_client.get_bucket_location.return_value = {'LocationConstraint': ''}
        provider = S3ClientProvider(session)
        client = provider.get_client('foo')
        self.assertEqual(1, session.create_client.call_count)
        self.assertEqual(client, provider.get_client('foo'))
        self.assertEqual(1, session.create_client.call_count)

    def test_removes_cli_error_events(self):
        # We should also remove the error handler for S3.
        # This can be removed once the client switchover is done.
        session = mock.Mock()
        s3_client = mock.Mock()
        session.create_client.return_value = s3_client
        s3_client.get_bucket_location.return_value = {'LocationConstraint': ''}
        provider = S3ClientProvider(session)
        provider.get_client('foo')



# Copyright 2022 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
from dataclasses import dataclass, field

from awscli_venv import AwsCliVenv
from constants import (
    DIST_INFO_DIRECTORIES_TO_KEEP,
    DISTRIBUTION_SOURCE_EXE,
    EXE_ASSETS_DIR,
    PYINSTALLER_DIR,
    PYINSTALLER_EXE_NAME,
)
from utils import Utils


@dataclass
class ExeBuilder:
    workspace: str
    venv: AwsCliVenv

    _exe_dir: str = field(init=False)
    _final_dist_dir: str = field(init=False)
    _dist_dir: str = field(init=False)
    _build_dir: str = field(init=False)

    _utils: Utils = field(default_factory=lambda: Utils())

    def __post_init__(self):
        self._exe_dir = os.path.join(self.workspace, "aws")
        self._final_dist_dir = os.path.join(self._exe_dir, "dist")
        self._dist_dir = os.path.join(self.workspace, "dist")
        self._build_dir = os.path.join(self.workspace, "build")

    def build(self, cleanup=True):
        self._ensure_no_existing_build_dir()
        self._build_aws()
        self._build_aws_completer()
        self._utils.copy_directory_contents_into(EXE_ASSETS_DIR, self._exe_dir)
        self._update_metadata()
        if cleanup:
            self._cleanup()
        print(f"Built exe at {self._exe_dir}")

    def _update_metadata(self):
        self._utils.update_metadata(
            self._final_dist_dir,
            distribution_source=DISTRIBUTION_SOURCE_EXE,
        )
        for distinfo in self._utils.glob(
            '**/*.dist-info', root=self._final_dist_dir
        ):
            if not any(
                package_name in distinfo
                for package_name in DIST_INFO_DIRECTORIES_TO_KEEP
            ):
                self._utils.rmtree(
                    os.path.join(self._final_dist_dir, distinfo)
                )

    def _ensure_no_existing_build_dir(self):
        if self._utils.isdir(self._dist_dir):
            self._utils.rmtree(self._dist_dir)

    def _build_aws(self):
        aws_exe_build_dir = self._run_pyinstaller("aws.spec")
        self._utils.copy_directory(aws_exe_build_dir, self._final_dist_dir)

    def _build_aws_completer(self):
        aws_completer_exe_build_dir = self._run_pyinstaller(
            "aws_completer.spec"
        )
        self._utils.copy_directory_contents_into(
            aws_completer_exe_build_dir, self._final_dist_dir
        )

    def _run_pyinstaller(self, specfile: str):
        aws_spec_path = os.path.join(PYINSTALLER_DIR, specfile)
        self._utils.run(
            [
                self.venv.python_exe,
                os.path.join(self.venv.bin_dir, PYINSTALLER_EXE_NAME),
                aws_spec_path,
                "--distpath",
                self._dist_dir,
                "--workpath",
                self._build_dir,
            ],
            cwd=PYINSTALLER_DIR,
            check=True,
        )
        return os.path.join(
            self.workspace, "dist", os.path.splitext(specfile)[0]
        )

    def _cleanup(self):
        locations = [
            self._build_dir,
            self._dist_dir,
        ]
        for location in locations:
            self._utils.rmtree(location)
            print(f"Deleted build directory: {location}")
